from collections.abc import Iterable
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from frappe_manager import STABLE_APP_BRANCH_MAPPING_LIST
from frappe_manager.docker_wrapper.DockerException import DockerException
from frappe_manager.docker_wrapper.subprocess_output import SubprocessOutput
from frappe_manager.site_manager.site_exceptions import (
    BenchOperationBenchAppInSiteFailed,
    BenchOperationBenchBuildFailed,
    BenchOperationBenchInstallAppInPythonEnvFailed,
    BenchOperationBenchRemoveAppFromPythonEnvFailed,
    BenchOperationBenchSiteCreateFailed,
    BenchOperationException,
    BenchOperationFrappeBranchChangeFailed,
    BenchOperationRequiredDockerImagesNotAvailable,
    BenchOperationWaitForRequiredServiceFailed,
)
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.utils.docker import parameters_to_options
from frappe_manager.utils.site import get_all_docker_images


class BenchOperations:
    def __init__(self, bench) -> None:
        self.bench = bench
        self.bench_cli_cmd = ["/opt/.pyenv/shims/bench"]
        self.frappe_bench_dir = self.bench.path / "workspace" / "frappe-bench"

    def create_fm_bench(self):

        richprint.change_head("Configuring common_site_config.json")
        common_site_config_data = self.bench.bench_config.get_commmon_site_config_data(
            self.bench.services.database_manager.database_server_info
        )
        self.bench.common_bench_config_set(common_site_config_data)
        richprint.print("Configured common_site_config.json")

        richprint.change_head("Configuring frappe server")
        self.setup_frappe_server_config()
        richprint.print("Configured frappe server")

        self.setup_supervisor(force=True)

        self.change_frappeverse_prebaked_app_branch(app="frappe", branch=self.bench.bench_config.frappe_branch)

        self.is_required_services_available()

        self.bench_install_apps(self.bench.bench_config.apps_list)

        self.frappe_container_run(
            "rm -rf /workspace/frappe-bench/archived",
            BenchOperationException(self.bench.name, "Failed to remove /workspace/frappe-bench/archived directory."),
        )

        richprint.change_head(f"Creating bench site {self.bench.name}")
        self.create_bench_site()
        richprint.print(f"Created bench site {self.bench.name}")

        self.bench_install_apps_site()

    def create_bench_site(self):
        new_site_command = self.bench_cli_cmd + ["new-site"]
        new_site_command += ["--db-root-password", self.bench.services.database_manager.database_server_info.password]
        new_site_command += ["--db-name", self.bench.bench_config.db_name]
        new_site_command += ["--db-host", self.bench.services.database_manager.database_server_info.host]
        new_site_command += ["--admin-password", self.bench.bench_config.admin_pass]
        new_site_command += ["--db-port", str(self.bench.services.database_manager.database_server_info.port)]
        new_site_command += ["--verbose", "--no-mariadb-socket"]
        new_site_command += [self.bench.name]

        new_site_command = " ".join(new_site_command)

        self.frappe_container_run(
            new_site_command, raise_exception_obj=BenchOperationBenchSiteCreateFailed(self.bench.name)
        )

        self.frappe_container_run(
            " ".join(self.bench_cli_cmd + [f"use {self.bench.name}"]),
            raise_exception_obj=BenchOperationException(
                self.bench.name, f"Failed to set {self.bench.name} as default site."
            ),
        )

        self.frappe_container_run(
            " ".join(self.bench_cli_cmd + [f"--site {self.bench.name} scheduler enable"]),
            raise_exception_obj=BenchOperationException(
                self.bench.name, f"Failed to enable {self.bench.name}'s scheduler."
            ),
        )

    def is_required_services_available(self):
        richprint.change_head("Checking if required services are available.")
        required_services = {
            self.bench.services.database_manager.database_server_info.host: self.bench.services.database_manager.database_server_info.port,
            f"{self.bench.bench_config.container_name_prefix}-redis-cache": 6379,
            f"{self.bench.bench_config.container_name_prefix}-redis-queue": 6379,
            f"{self.bench.bench_config.container_name_prefix}-redis-socketio": 6379,
        }
        for service, port in required_services.items():
            output: SubprocessOutput = self.wait_for_required_service(host=service, port=port)
            if output.combined:
                richprint.print(output.combined[-1].replace('wait-for-it: ', ''), highlight=False)

    def frappe_container_run(
        self,
        command: str,
        raise_exception_obj: Optional[BenchOperationException] = None,
        capture_output: bool = False,
        user: str = "frappe",
        workdir="/workspace/frappe-bench",
    ):
        try:
            if capture_output:
                output: SubprocessOutput = self.bench.compose_project.docker.compose.exec(
                    service="frappe", command=command, user=user, workdir=workdir, stream=not capture_output
                )
                return output
            else:
                output: Iterable[Tuple[str, bytes]] = self.bench.compose_project.docker.compose.exec(
                    service="frappe", command=command, workdir=workdir, user=user, stream=not capture_output
                )
                richprint.live_lines(output)

        except DockerException as e:
            if raise_exception_obj:
                raise_exception_obj.set_output(e.output)
                raise raise_exception_obj
            raise e

    def change_frappeverse_prebaked_app_branch(self, app: str, branch: str):
        richprint.change_head(f"Configuring {app} app's branch -> {branch}")

        if not branch == STABLE_APP_BRANCH_MAPPING_LIST[app]:
            richprint.change_head(
                f"Changing prebaked {app} app's branch {STABLE_APP_BRANCH_MAPPING_LIST[app]} -> {self.bench.bench_config.frappe_branch}"
            )
            change_frappe_branch_command = self.bench_cli_cmd + [f"get-app --overwrite --branch {branch} frappe"]
            change_frappe_branch_command = " ".join(change_frappe_branch_command)

            exception = BenchOperationFrappeBranchChangeFailed(
                bench_name=self.bench.name, app=app, branch=self.bench.bench_config.frappe_branch
            )

            self.frappe_container_run(command=change_frappe_branch_command, raise_exception_obj=exception)

        richprint.print(f"Configured {app} app's branch -> {self.bench.bench_config.frappe_branch}")

    def setup_supervisor(self, force: bool = False):
        config_dir_path: Path = self.frappe_bench_dir / "config"
        supervisor_conf_path: Path = config_dir_path / "supervisor.conf"

        richprint.change_head("Checking supervisor configuration")
        if not supervisor_conf_path.exists() or force:
            richprint.change_head("Configuring supervisor configs")

            bench_setup_supervisor_command = self.bench_cli_cmd + [
                "setup supervisor --skip-redis --skip-supervisord --yes --user frappe"
            ]

            bench_setup_supervisor_command = " ".join(bench_setup_supervisor_command)
            bench_setup_supervisor_exception = BenchOperationException(
                self.bench.name, "Failed to configure supervisor."
            )
            self.frappe_container_run(bench_setup_supervisor_command, bench_setup_supervisor_exception)
            self.split_supervisor_config()
            richprint.print("Configured supervisor configs")

    def split_supervisor_config(self):
        import configparser

        supervisor_conf_path: Path = self.frappe_bench_dir / "config" / "supervisor.conf"
        config = configparser.ConfigParser(allow_no_value=True, strict=False, interpolation=None)
        config.read_string(supervisor_conf_path.read_text())

        for section_name in config.sections():
            if "group:" not in section_name:
                section_config = configparser.ConfigParser(interpolation=None)
                section_config.add_section(section_name)
                for key, value in config.items(section_name):
                    if "frappe-bench-frappe-web" in section_name:
                        if key == "command":
                            value = value.replace("127.0.0.1:80", "0.0.0.0:80")
                    section_config.set(section_name, key, value)
                if "worker" in section_name:
                    file_name = f"{section_name.replace('program:','')}.workers.fm.supervisor.conf"
                else:
                    file_name = f"{section_name.replace('program:','')}.fm.supervisor.conf"

                new_file: Path = supervisor_conf_path.parent / file_name

                with open(new_file, "w") as section_file:
                    section_config.write(section_file)

                self.bench.logger.info(f"Split supervisor conf {section_name} => {file_name}")

    def setup_frappe_server_config(self):
        bench_serve_help_output: Optional[SubprocessOutput] = self.frappe_container_run(
            " ".join(self.bench_cli_cmd + ["serve --help"]), capture_output=True
        )
        bench_dev_server_script_output = self.frappe_container_run(
            "cat /opt/user/bench-dev-server", capture_output=True
        )
        import re

        if "host" in " ".join(bench_serve_help_output.combined):
            new_bench_dev_server_script = re.sub(
                r"--port \d+", "--host 0.0.0.0 --port 80", " ".join(bench_dev_server_script_output.combined)
            )
        else:
            new_bench_dev_server_script = re.sub(
                r"--port \d+", "--port 80", " ".join(bench_dev_server_script_output.combined)
            )

        self.frappe_container_run(f'echo "{new_bench_dev_server_script}" > /opt/user/bench-dev-server.sh')
        self.frappe_container_run("chmod +x /opt/user/bench-dev-server.sh", user='root')

    def bench_install_apps(self, apps_lists, already_installed_apps: Dict = STABLE_APP_BRANCH_MAPPING_LIST):
        to_install_apps = [x["app"] for x in apps_lists]

        for app, branch in already_installed_apps.items():
            if app == 'frappe':
                continue

            if app not in to_install_apps:
                richprint.change_head(f"Removing prebaked app {app} from python env.")
                self.bench_rm_app_env(app)
                richprint.print(f"Removed prebaked app {app}")

        for app_info in apps_lists:
            app = app_info["app"]
            branch = app_info["branch"]

            status_txt = f"Building and Installing app {app} in env."

            if branch:
                status_txt = f"Building and Installing app {app} -> {branch}."

            richprint.change_head(status_txt)

            if app in already_installed_apps.keys():
                if already_installed_apps[app] == branch:
                    richprint.print(f"Skipped installation of prebaked app [blue]{app} -> {branch}[/blue].")
                    continue

                if not branch:
                    branch = already_installed_apps[app]

            self.bench_install_app_env(app, branch)

            richprint.print(f"Builded and Installed app [blue]{app}{' -> ' + branch if branch else ''}[/blue] in env.")

    def bench_install_apps_site(self):
        apps_dir = self.frappe_bench_dir / 'apps'
        apps_dirs: List[Path] = [item for item in apps_dir.iterdir() if item.is_dir()]

        for app in apps_dirs:
            richprint.change_head(f"Installing app {app.name} in site.")
            self.bench_install_app_site(app.name)
            richprint.print(f"Installed app {app.name} in site.")

    def bench_build(self, app_list: Optional[List[str]] = None):
        build_cmd = self.bench_cli_cmd + ["build"]

        if app_list is not None:
            for app in app_list:
                build_cmd += ["--app"] + [app]

        build_exception = BenchOperationBenchBuildFailed(bench_name=self.bench.name, apps=app_list)

        build_cmd = " ".join(build_cmd)
        self.frappe_container_run(build_cmd, build_exception)

    def bench_install_app_env(
        self, app: str, branch: Optional[str] = None, overwrite: bool = True, skip_assets: bool = False
    ):
        parameters: Dict = locals()

        remove_parameters = ["app"]
        app_install_env_command = self.bench_cli_cmd + ["get-app"]
        app_install_env_command += parameters_to_options(parameters, exclude=remove_parameters)
        app_install_env_command += [app]

        app_install_env_command = " ".join(app_install_env_command)
        app_install_exception = BenchOperationBenchInstallAppInPythonEnvFailed(bench_name=self.bench.name, app_name=app)

        self.frappe_container_run(
            app_install_env_command,
            raise_exception_obj=app_install_exception,
        )

    def bench_rm_app_env(self, app: str, no_backup: bool = True, force: bool = True):
        parameters: dict = locals()

        remove_parameters = ["app"]
        app_rm_env_command = self.bench_cli_cmd + ["rm"]
        app_rm_env_command += parameters_to_options(parameters, exclude=remove_parameters)
        app_rm_env_command += [app]

        app_rm_env_command = " ".join(app_rm_env_command)

        self.frappe_container_run(
            app_rm_env_command,
            raise_exception_obj=BenchOperationBenchRemoveAppFromPythonEnvFailed(
                bench_name=self.bench.name, app_name=app
            ),
        )

    def bench_install_app_site(self, app: str):
        app_install_site_command = self.bench_cli_cmd + ["--site", self.bench.name]
        app_install_site_command += ["install-app", app]
        app_install_site_command = " ".join(app_install_site_command)

        self.frappe_container_run(
            app_install_site_command,
            raise_exception_obj=BenchOperationBenchAppInSiteFailed(bench_name=self.bench.name, app_name=app),
        )

    def is_bench_site_exists(self, bench_site_name: str):
        site_path: Path = self.frappe_bench_dir / "sites" / bench_site_name
        return site_path.exists()

    def wait_for_required_service(self, host: str, port: int, timeout: int = 120):
        return self.frappe_container_run(
            f"wait-for-it -t {timeout} {host}:{port}",
            raise_exception_obj=BenchOperationWaitForRequiredServiceFailed(
                bench_name=self.bench.name, host=host, port=port, timeout=timeout
            ),
            capture_output=True,
        )

    def check_required_docker_images_available(self):
        richprint.change_head("Checking required docker images availability")
        fm_images = get_all_docker_images()
        system_available_images = self.bench.compose_project.docker.images()

        not_available_images = []

        for key, value in fm_images.items():
            name = value['name']
            tag = value['tag']

            found = False

            for item in system_available_images:
                if item.get('Repository') == name and item.get('Tag') == tag:
                    found = True
                    break

            if not found:
                image = f"{name}:{tag}"
                not_available_images.append(image)

        # remove duplicates
        not_available_images = list(dict.fromkeys(not_available_images))

        if not_available_images:
            for image in not_available_images:
                richprint.error(f"Docker image '{image}' is not available locally")
            raise BenchOperationRequiredDockerImagesNotAvailable(self.bench.name, 'fm self update images')



from pydantic import BaseModel
from typing import List, Optional
from frappe_manager.ssl_manager import SUPPORTED_SSL_TYPES
from frappe_manager.utils.site import is_wildcard_fqdn


class SSLCertificate(BaseModel):
    domain: str
    ssl_type: SUPPORTED_SSL_TYPES
    hsts: str = 'off'
    alias_domains: List[str] = []
    toml_exclude: Optional[set] = {'domain', 'alias_domains', 'toml_exclude'}

    @property
    def has_wildcard(self) -> bool:
        return any(is_wildcard_fqdn(domain) for domain in self.alias_domains)



import json
from pathlib import Path
import time
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.docker_wrapper.DockerException import DockerException
from frappe_manager.site_manager.site_exceptions import AdminToolsFailedToStart, BenchException
from frappe_manager.ssl_manager.nginxproxymanager import NginxProxyManager
from frappe_manager.utils.helpers import get_container_name_prefix, get_current_fm_version, get_template_path


class AdminTools:
    def __init__(self, bench_name: str, bench_path: Path, nginx_proxy: NginxProxyManager, verbose: bool = True):
        self.compose_path = bench_path / "docker-compose.admin-tools.yml"
        self.bench_name = bench_name
        self.quiet = not verbose
        self.compose_project = ComposeProject(
            ComposeFile(self.compose_path, template_name='docker-compose.admin-tools.tmpl')
        )
        self.nginx_proxy: NginxProxyManager = nginx_proxy
        self.nginx_config_location_path: Path = self.nginx_proxy.dirs.conf.host / 'custom' / 'admin-tools.conf'

    def generate_compose(self, db_host: str):
        # env set db_host for ADMINER_DEFAULT_SERVER
        env = {"ADMINER_DEFAULT_SERVER": db_host}
        self.compose_project.compose_file_manager.set_envs('adminer', env)

        self.compose_project.compose_file_manager.yml = self.compose_project.compose_file_manager.load_template()
        self.compose_project.compose_file_manager.set_container_names(get_container_name_prefix(self.bench_name))
        self.compose_project.compose_file_manager.yml["networks"]["site-network"]["name"] = (
            self.bench_name.replace(".", "") + f"-network"
        )
        self.compose_project.compose_file_manager.set_version(get_current_fm_version())
        self.compose_project.compose_file_manager.write_to_file()

    def create(self, db_host: str):
        richprint.change_head("Generating admin tools configuration")
        self.generate_compose(db_host)
        richprint.print("Generating admin tools configuration: Done")

    def save_nginx_location_config(self):
        data = {
            "mailhog_host": f"{get_container_name_prefix(self.bench_name)}-mailhog",
            "adminer_host": f"{get_container_name_prefix(self.bench_name)}-adminer",
        }
        from jinja2 import Template

        template_path: Path = get_template_path('admin-tools-location.tmpl')

        template = Template(template_path.read_text())
        output = template.render(data)

        if not self.nginx_config_location_path.parent.exists():
            self.nginx_config_location_path.mkdir(exist_ok=True)

        self.nginx_config_location_path.write_text(output)

    def remove_nginx_location_config(self):
        if self.nginx_config_location_path.exists():
            self.nginx_config_location_path.unlink()

    def get_common_site_config(self, common_bench_config_path: Path):
        if not common_bench_config_path.exists():
            raise BenchException(self.bench_name, message='common_site_config.json not found.')

        current_common_site_config = json.loads(common_bench_config_path.read_bytes())

        return current_common_site_config

    def configure_mailhog_as_default_server(self, force: bool = False):
        richprint.change_head("Configuring Mailhog as default mail server.")
        common_bench_config_path = self.compose_path.parent / "workspace/frappe-bench/sites/common_site_config.json"

        current_common_site_config = self.get_common_site_config(common_bench_config_path)

        new_conf = {
            "mail_port": 1025,
            "mail_server": f"{get_container_name_prefix(self.bench_name)}-mailhog",
            "disable_mail_smtp_authentication": 1,
        }

        if not force:
            if 'mail_server' in current_common_site_config:
                if not current_common_site_config['mail_server'] == new_conf['mail_server']:
                    raise BenchException(
                        self.bench_name,
                        'Failed to set MailHog as the default mail server because another mail server is currently configured as the default.',
                    )

        frappe_server_restart_required = False

        for key, value in new_conf.items():
            if key not in current_common_site_config:
                current_common_site_config[key] = value
                frappe_server_restart_required = True

            elif not current_common_site_config[key] == value:
                current_common_site_config[key] = value
                frappe_server_restart_required = True

        common_bench_config_path.write_text(json.dumps(current_common_site_config))
        richprint.print("Configured Mailhog as default mail server.")
        return frappe_server_restart_required

    def remove_mailhog_as_default_server(self):
        richprint.change_head("Removing Mailhog as default mail server.")
        common_bench_config_path = self.compose_path.parent / "workspace/frappe-bench/sites/common_site_config.json"
        current_common_site_config = self.get_common_site_config(common_bench_config_path)

        new_conf = {
            "mail_port": 1025,
            "mail_server": f"{get_container_name_prefix(self.bench_name)}-mailhog",
            "disable_mail_smtp_authentication": 1,
        }

        frappe_server_restart_required = False
        for key, value in new_conf.items():
            if not key in current_common_site_config:
                continue
            if not current_common_site_config[key] == value:
                continue
            del current_common_site_config[key]
            frappe_server_restart_required = True

        common_bench_config_path.write_text(json.dumps(current_common_site_config))
        richprint.print("Removed Mailhog as default mail server.")
        return frappe_server_restart_required

    def wait_till_services_started(self, interval=2, timeout=30):
        admin_tools_services = ['mailhog:8025', 'adminer:8080']

        for tool in admin_tools_services:
            running = False
            for i in range(timeout):
                try:
                    check_command = f"wait-for-it -t {interval} {get_container_name_prefix(self.bench_name)}-{tool}"
                    self.nginx_proxy.compose_project.docker.compose.exec(
                        service='nginx', command=check_command, stream=False
                    )

                    running = True
                    break
                except DockerException as e:
                    continue

            if not running:
                raise AdminToolsFailedToStart(self.bench_name)

    def enable(self, force_recreate_container: bool = False, force_configure: bool = False) -> bool:
        self.compose_project.start_service(force_recreate=force_recreate_container)
        self.wait_till_services_started()
        self.save_nginx_location_config()
        self.nginx_proxy.reload()
        frappe_server_restart_required = self.configure_mailhog_as_default_server(force=force_configure)
        return frappe_server_restart_required

    def disable(self) -> bool:
        self.compose_project.stop_service()
        self.remove_nginx_location_config()
        self.nginx_proxy.reload()
        frappe_server_restart_required = self.remove_mailhog_as_default_server()
        return frappe_server_restart_required



from pathlib import Path
import tomlkit
from frappe_manager.migration_manager.migration_base import MigrationBase
from frappe_manager.migration_manager.migration_helpers import MigrationBenches, MigrationServicesManager
from frappe_manager.migration_manager.version import Version
from frappe_manager.migration_manager.backup_manager import BackupManager
from frappe_manager.display_manager.DisplayManager import richprint


class MigrationV0140(MigrationBase):
    version = Version("0.14.0")

    def init(self):
        self.cli_dir: Path = Path.home() / 'frappe'
        self.benches_dir = self.cli_dir / "sites"
        self.backup_manager = BackupManager(name=str(self.version), benches_dir=self.benches_dir)
        self.benches_manager = MigrationBenches(self.benches_dir)
        self.services_manager: MigrationServicesManager = MigrationServicesManager(
            services_path=self.cli_dir / 'services'
        )

    def migrate_services(self):
        richprint.change_head("Migrating [blue]fm_config.toml[/blue] changes")
        config_path = self.cli_dir / 'fm_config.toml'

        if config_path.exists():
            data = tomlkit.parse(config_path.read_text())
            email = data.get('le_email', None)

            if email:
                data['letsencrypt'] = {'email': email}
                del data['le_email']

            with open(config_path, 'w') as f:
                f.write(tomlkit.dumps(data))
        richprint.print("Migrated [blue]fm_config.toml[/blue]")



import shutil
from typing import Optional
from frappe_manager.migration_manager.migration_helpers import MigrationBench
from rich.padding import Padding
from rich.text import Text
import importlib
import pkgutil
from pathlib import Path
from frappe_manager import CLI_DIR, CLI_SITES_ARCHIVE
from frappe_manager.metadata_manager import FMConfigManager
from frappe_manager.migration_manager.migration_exections import (
    MigrationExceptionInBench,
)
from frappe_manager.utils.helpers import capture_and_format_exception, install_package, get_current_fm_version
from frappe_manager.logger import log
from frappe_manager.migration_manager.version import Version
from frappe_manager.display_manager.DisplayManager import richprint


class MigrationExecutor:
    """
    Migration executor class.

    This class is responsible for executing migrations.
    """

    def __init__(self, fm_config_manager: FMConfigManager):
        self.fm_config_manager: FMConfigManager = fm_config_manager
        self.prev_version = self.fm_config_manager.version
        self.rollback_version = self.fm_config_manager.version
        self.current_version = Version(get_current_fm_version())
        self.migrations_path = Path(__file__).parent / "migrations"
        self.logger = log.get_logger()
        self.migrations = []
        self.undo_stack = []
        self.migrate_benches = {}

    def execute(self):
        """
        Execute the migration.
        This method will execute the migration and return the number of
        executed statements.
        """

        if not self.prev_version < self.current_version:
            return True

        current_migration = None

        # Dynamically import all modules in the 'migrations' subfolder
        for _, name, _ in pkgutil.iter_modules([str(self.migrations_path)]):
            try:
                module = importlib.import_module(f".migrations.{name}", __package__)
                for attr_name in dir(module):
                    attr = getattr(module, attr_name)
                    if (
                        isinstance(attr, type)
                        and hasattr(attr, "up")
                        and hasattr(attr, "down")
                        and hasattr(attr, "set_migration_executor")
                        and hasattr(attr, "version")
                    ):
                        if not getattr(attr, "version") == Version('0.0.0'):
                            migration = attr()
                            migration.set_migration_executor(migration_executor=self)
                            current_migration = migration

                            if migration.version > self.prev_version and migration.version <= self.current_version:
                                self.migrations.append(migration)

            except Exception as e:
                exception_str = capture_and_format_exception()
                print(f"Failed to register migration {name}: {exception_str}")

        self.migrations = sorted(self.migrations, key=lambda x: x.version)

        if self.migrations:
            richprint.print("Pending Migrations...", emoji_code=':counterclockwise_arrows_button:')

            for migration in self.migrations:
                richprint.print(f"[bold]v{migration.version}[/bold]", emoji_code=':package:')

            richprint.print("This process may take a while.", emoji_code="\n:hourglass_not_done:")

            richprint.print(
                "For a manual migration guide, visit https://github.com/rtCamp/Frappe-Manager/wiki/Migrations#manual-migration-procedure",
                emoji_code=":blue_book:",
            )

            migrate_msg = [
                "\nOptions :\n",
                "[blue]\[yes][/blue] Start Migration: Proceed with the migration process.",
                "[blue]\[no][/blue]  Abort and Revert: Do not migrate and revert to the previous fm version.",
                "\nDo you want to proceed with the migration ?",
            ]
            continue_migration = richprint.prompt_ask(prompt="\n".join(migrate_msg), choices=["yes", "no"])

            if continue_migration == "no":
                install_package("frappe-manager", str(self.prev_version.version))
                richprint.exit(
                    f"Successfully installed [bold][blue]Frappe-Manager[/blue][/bold] version: v{str(self.prev_version.version)}",
                    emoji_code=":white_check_mark:",
                )

        rollback = False
        archive = False

        exception_migration_in_bench_occured = False
        try:
            # run all the migrations
            prev_migration = None
            for migration in self.migrations:
                richprint.change_head(f"Running migration introduced in v{migration.version}")
                self.logger.info(f"[{migration.version}] : Migration starting")
                try:
                    self.undo_stack.append(migration)

                    migration.up()

                    prev_migration = migration
                    if not exception_migration_in_bench_occured:
                        self.rollback_version = prev_migration.get_rollback_version()

                except MigrationExceptionInBench as e:
                    captured_output = capture_and_format_exception(traceback_max_frames=0)
                    self.logger.error(f"[{migration.version}] : Migration Failed\n{captured_output}")

                    if migration.version < self.migrations[-1].version:
                        exception_migration_in_bench_occured = True
                        continue
                    raise e

                except Exception as e:
                    captured_output = capture_and_format_exception()
                    self.logger.error(f"[{migration.version}] : Migration Failed\n{captured_output}")
                    raise e

            if exception_migration_in_bench_occured:
                raise MigrationExceptionInBench('')

        except MigrationExceptionInBench as e:
            if self.migrate_benches:
                passed_print_head = True

                for bench, bench_status in self.migrate_benches.items():
                    if not bench_status["exception"]:
                        if passed_print_head:
                            richprint.stdout.rule('[bold]Migration Passed Benches[bold]', style='green')
                            passed_print_head = False

                        richprint.print(f"[green]Bench[/green]: {bench}", emoji_code=':construction:')

                failed_print_head = True

                for bench, bench_status in self.migrate_benches.items():
                    if bench_status["exception"]:
                        if failed_print_head:
                            richprint.stdout.rule(
                                ':police_car_light: [bold][red]Migration Failed Benches[/red][bold] :police_car_light:',
                                style='red',
                            )
                            failed_print_head = False

                        richprint.error(f"[red]Bench[/red]: {bench}", emoji_code=':construction:')

                        richprint.error(
                            f"[red]Failed Migration Version[/red]: {bench_status['last_migration_version']}",
                            emoji_code=':package:',
                        )

                        richprint.error(
                            f"[red]Exception[/red]: {type(bench_status['exception']).__name__}",
                            emoji_code=':stop_sign:',
                        )
                        richprint.stdout.print(Padding(Text(text=str(bench_status['exception'])), (0, 0, 0, 3)))

                richprint.print(f"For error specifics, refer to {CLI_DIR}/logs/fm.log", emoji_code=':page_facing_up:')

                if not failed_print_head:
                    richprint.stdout.rule(style='red')
                else:
                    if not passed_print_head:
                        richprint.stdout.rule(style='green')

                archive_msg = [
                    'Available options after migrations failure :',
                    f"[blue]\[yes][/blue] Archive failed benches : Benches that have failed will be rolled back to there last successfully completed migration version and stored in '{CLI_SITES_ARCHIVE}'.",
                    '[blue]\[no][/blue] Revert migration : Restore the FM CLI and FM environment to the last successfully completed migration version for all benches.',
                    '\nDo you wish to archive all benches that failed during migration ?',
                ]
                archive = richprint.prompt_ask(prompt="\n".join(archive_msg), choices=["yes", "no"])

                if archive == "no":
                    rollback = True

        except Exception as e:
            richprint.error(f"[red]Migration failed[red] : {e}")
            rollback = True

        if archive == "yes":
            self.prev_version = self.undo_stack[-1].version
            for bench, bench_info in self.migrate_benches.items():
                if bench_info["exception"]:
                    archive_bench_path = CLI_SITES_ARCHIVE / bench
                    CLI_SITES_ARCHIVE.mkdir(exist_ok=True, parents=True)
                    shutil.move(bench_info["object"].path, archive_bench_path)
                    richprint.print(f"[bold]Archived bench :[/bold] [yellow]{bench}[/yellow]")

        if rollback:
            self.rollback()
            self.fm_config_manager.version = self.rollback_version
            self.fm_config_manager.export_to_toml()
            richprint.print(
                f"Installing [bold][blue]Frappe-Manager[/blue][/bold] version: v{str(self.rollback_version.version)}"
            )
            install_package("frappe-manager", str(self.rollback_version.version))
            richprint.exit("Rollback complete.", emoji_code=':back:')

        self.fm_config_manager.version = self.current_version
        self.fm_config_manager.export_to_toml()
        return True

    def set_bench_data(
        self,
        bench: MigrationBench,
        exception=None,
        migration_version: Optional[Version] = None,
        traceback_str: Optional[str] = None,
    ):
        self.migrate_benches[bench.name] = {
            "object": bench,
            "exception": exception,
            "last_migration_version": migration_version,
            "traceback": traceback_str,
        }

    def get_site_data(self, bench_name):
        try:
            data = self.migrate_benches[bench_name]
        except KeyError as e:
            return None
        return data

    def rollback(self):
        """
        Rollback the migration.
        This method will rollback the migration and return the number of
        rolled back statements.
        """

        # run all the migrations
        for migration in reversed(self.undo_stack):
            if migration.version > self.rollback_version:
                richprint.change_head(f"Rolling back migration introduced in v{migration.version}")
                self.logger.info(f"[{migration.version}] : Rollback starting")
                try:
                    migration.down()
                except Exception as e:
                    self.logger.error(f"[{migration.version}] : Rollback Failed\n{e}")
                    raise e



from rich.console import Console, Group
from rich.prompt import Prompt
from rich.style import Style
from rich.theme import Theme
from rich.spinner import Spinner
from rich.live import Live
from rich.text import Text
from rich.padding import Padding
from rich.table import Table

import typer
from collections import deque
from typing import Optional

error = Style()
theme = Theme({"errors": error})


class DisplayManager:
    def __init__(self):
        self.stdout = Console()
        # self.stderr = Console(stderr=True)
        self.previous_head = None
        self.current_head = None
        self.spinner = Spinner(text=self.current_head, name="dots2", speed=1)
        self.live = Live(self.spinner, console=self.stdout, transient=True)

    def start(self, text: str):
        """
        Starts the display manager with the given text.

        Args:
            text (str): The text to be displayed.

        Returns:
            None
        """
        self.current_head = self.previous_head = Text(text=text, style="bold blue")
        self.spinner = Spinner(text=self.current_head, name="dots2", speed=1)
        self.live.start(refresh=True)
        self.live.update(self.spinner, refresh=True)

    def error(self, text: str, exception: Optional[Exception] = None, emoji_code: str = ":x:"):
        """
        Display an error message with an optional emoji code.

        Args:
            text (str): The error message to display.
            emoji_code (str, optional): The emoji code to display before the error message. Defaults to ':x:'.
        """
        self.stdout.print(f"{emoji_code} {text}")

        if exception:
            raise exception

    def warning(self, text: str, emoji_code: str = ":warning: "):
        """
        Display a warning message with an optional emoji code.

        Args:
            text (str): The warning message to display.
            emoji_code (str, optional): The emoji code to prepend to the message. Defaults to ":warning: ".

        Returns:
            None
        """
        self.stdout.print(f"{emoji_code} {text}")

    def exit(self, text: str, emoji_code: str = ":x:", os_exit=False, error_msg=None):
        """
        Exits the display manager and prints the given text with an optional emoji code and error message.

        Args:
            text (str): The text to be printed.
            emoji_code (str, optional): The emoji code to be displayed before the text. Default is ":x:".
            os_exit (bool, optional): If True, the program will exit with status code 1. Default is False.
            error_msg (str, optional): The error message to be displayed after the text. Default is None.
        """
        self.stop()

        to_print = f"{emoji_code} {text}"
        if error_msg:
            to_print = f"{emoji_code} {text}\n Error : {error_msg}"

        self.stdout.print(to_print)

        if os_exit:
            exit(1)

        raise typer.Exit(1)

    def print(self, text: str, emoji_code: str = ":white_check_mark:", prefix: Optional[str] = None, **kwargs):
        """
        Prints the given text with an optional emoji code.

        Args:
            text (str): The text to be printed.
            emoji_code (str, optional): The emoji code to be displayed before the text. Defaults to ":white_check_mark:".
        """
        msg = f"{emoji_code} {text}"

        if prefix:
            msg = f"{emoji_code} {prefix} {text}"

        self.stdout.print(msg, **kwargs)

    def update_head(self, text: str):
        """
        Update the head of the display manager with the given text.

        Args:
            text (str): The new head text.

        Returns:
            None
        """
        self.previous_head = self.current_head
        self.current_head = text
        self.live.console.print(self.previous_head, style="blue")
        self.spinner.update(text=Text(self.current_head, style="blue bold"), style="bold blue")

    def prompt_ask(self, **args):
        self.spinner.update()
        self.live.stop()
        value = Prompt.ask(**args, console=self.stdout)
        self.start('Working')
        return value

    def change_head(self, text: str, style: Optional[str] = 'blue bold'):
        """
        Change the head text and update the spinner and live display.

        Args:
            text (str): The new head text.

        Returns:
            None
        """
        self.previous_head = self.current_head
        self.current_head = text
        if style:
            self.spinner.update(text=Text(self.current_head, style="blue bold"))
        else:
            self.spinner.update(text=self.current_head)
        self.live.refresh()

    def update_live(self, renderable=None, padding: tuple = (0, 0, 0, 0)):
        """
        Update the live display with the given renderable object and padding.

        Args:
            renderable: The object to be rendered on the live display.
            padding: The padding values for the renderable object (top, right, bottom, left).
        """
        if renderable:
            if padding:
                renderable = Padding(renderable, padding)

            group = Group(self.spinner, renderable)
            self.live.update(group)
        else:
            self.live.update(self.spinner)
            self.live.refresh()

    def live_lines(
        self,
        data,
        stdout: bool = True,
        stderr: bool = True,
        lines: int = 4,
        padding: tuple = (0, 0, 0, 2),
        stop_string: Optional[str] = None,
        log_prefix: str = "=>",
    ):
        """
        Display live lines from the given data source.

        Args:
            data: An iterator that yields tuples of (source, line) where source is either "stdout" or "stderr" and line is a string.
            stdout: Whether to display lines from the stdout source. Default is True.
            stderr: Whether to display lines from the stderr source. Default is True.
            lines: The maximum number of lines to display. Default is 4.
            padding: A tuple of four integers representing the padding (top, right, bottom, left) around the displayed lines. Default is (0, 0, 0, 0).
            stop_string: A string that, if found in a line, will stop the display and return 0. Default is None.
            log_prefix: The prefix to add to each displayed line. Default is "=>".
            return_exit_code: Whether to return the exit code when stop_string is found. Default is False.
        """
        max_height = lines
        displayed_lines = deque(maxlen=max_height)

        while True:
            try:
                source, line = next(data)
                line = line.decode()
                # print(' --',line)

                if "[==".lower() in line.lower() or 'Updating files:'.lower() in line.lower():
                    continue

                if source == "stdout" and stdout:
                    displayed_lines.append(line)

                if source == "stderr" and stderr:
                    displayed_lines.append(line)

                if stop_string and stop_string.lower() in line.lower():
                    raise StopIteration

                table = Table(show_header=False, box=None)
                table.add_column()

                for linex in list(displayed_lines):
                    prefix_text = Text(log_prefix + ' ', no_wrap=True)
                    table_line = Text.from_ansi(linex)
                    prefix_text.append_text(table_line)
                    table.add_row(prefix_text)

                self.update_live(table, padding=padding)
                self.live.refresh()

            except KeyboardInterrupt:
                richprint.live.refresh()

            except StopIteration:
                self.update_live()
                break

    def stop(self):
        self.spinner.update()
        self.live.update(Text("", end=""))
        self.live.stop()


richprint = DisplayManager()



import typer
from frappe_manager.logger import log
from typing import List, Optional
from pathlib import Path
from rich.table import Table
from frappe_manager.services_manager.services import ServicesManager
from frappe_manager.site_manager.site import Bench
from frappe_manager.display_manager.DisplayManager import richprint


class BenchesManager:
    def __init__(self, sitesdir: Path, services: ServicesManager, verbose: bool = False):
        self.root_path = sitesdir
        self.benches: List[Bench] = []

        self.verbose = False
        self.services: ServicesManager = services
        self.logger = log.get_logger()

    def set_typer_context(self, ctx: typer.Context):
        self.ctx: typer.Context = ctx

    def get_all_bench(self, exclude: List[str] = []):
        sites = {}
        for dir in self.root_path.iterdir():
            if dir.is_dir() and dir.parts[-1] not in exclude:
                name = dir.parts[-1]
                dir = dir / "docker-compose.yml"
                if dir.exists():
                    sites[name] = dir
        return sites

    def add_bench(self, bench: Bench):
        self.benches.append(bench)

    def create_benches(self, is_template_bench: bool = False):
        for bench in self.benches:
            bench.create(is_template_bench=is_template_bench)

    def start_benches(self):
        for bench in self.benches:
            bench.start()

    def stop_benches(self):
        for bench in self.benches:
            bench.stop()

    def remove_benches(self):
        for bench in self.benches:
            bench.remove_bench()

    def list_benches(self):
        """
        Lists all the sites and their status.
        """

        # TODO entrypoint check can be changed
        richprint.change_head("Generating bench list")

        bench_list = self.get_all_bench()

        if not bench_list:
            richprint.exit(
                "Seems like you haven't created any sites yet. To create a bench, use the command: 'fm create <benchname>'.",
                emoji_code=":white_check_mark:",
            )

        list_table = Table(show_lines=True, show_header=True, highlight=True)
        list_table.add_column("Site")
        list_table.add_column("Status", vertical="middle")
        list_table.add_column("Path")

        for bench_name in bench_list.keys():
            try:
                bench = Bench.get_object(bench_name, self.services, workers_check=False, admin_tools_check=False)

                row_data = f"[link=http://{bench.name}]{bench.name}[/link]"
                path_data = f"[link=file://{bench.path}]{bench.path}[/link]"

                status_color = "white"
                status_msg = "Inactive"

                if bench.compose_project.running:
                    status_color = "green"
                    status_msg = "Active"

                status_data = f"[{status_color}]{status_msg}[/{status_color}]"

                list_table.add_row(row_data, status_data, path_data, style=f"{status_color}")
                richprint.update_live(list_table, padding=(0, 0, 0, 0))
            except FileNotFoundError as e:
                richprint.warning(f'[red][bold]{bench_name}[/bold][/red] : Bench config not found at {e.filename}')

        richprint.stop()

        if list_table.row_count:
            richprint.stdout.print(list_table)



import json
from typing import List
from rich.text import Text
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from frappe_manager.compose_project.exceptions import (
    DockerComposeProjectFailedToPullImagesError,
    DockerComposeProjectFailedToRemoveError,
    DockerComposeProjectFailedToRestartError,
    DockerComposeProjectFailedToStartError,
    DockerComposeProjectFailedToStopError,
)
from frappe_manager.docker_wrapper.DockerClient import DockerClient
from frappe_manager.docker_wrapper.DockerException import DockerException
from frappe_manager.display_manager.DisplayManager import richprint


class ComposeProject:
    def __init__(self, compose_file_manager: ComposeFile, verbose: bool = False):
        self.compose_file_manager: ComposeFile = compose_file_manager
        self.docker: DockerClient = DockerClient(compose_file_path=self.compose_file_manager.compose_path)
        self.quiet = not verbose

    def start_service(self, services: List[str] = [], force_recreate: bool = False):
        """
        Starts the specific compose service.
        """
        try:
            output = self.docker.compose.up(
                services=services, detach=True, pull="never", force_recreate=force_recreate, stream=self.quiet
            )
            if self.quiet:
                richprint.live_lines(output, padding=(0, 0, 0, 2))
        except DockerException as e:
            raise DockerComposeProjectFailedToStartError(self.compose_file_manager.compose_path, services)

    def stop_service(self, services: List[str] = [], timeout: int = 100):
        """
        Stops the specific compose service.
        """
        try:
            output = self.docker.compose.stop(services=services, timeout=timeout, stream=self.quiet)
            if self.quiet:
                richprint.live_lines(output, padding=(0, 0, 0, 2))
        except DockerException as e:
            raise DockerComposeProjectFailedToStopError(
                self.compose_file_manager.compose_path, self.compose_file_manager.get_services_list()
            )

    def down_service(self, remove_ophans=True, volumes=True, timeout=5):
        """
        Stops and removes the containers for the site.

        Args:
            remove_ophans (bool, optional): Whether to remove orphan containers. Defaults to True.
            volumes (bool, optional): Whether to remove volumes. Defaults to True.
            timeout (int, optional): Timeout in seconds for stopping the containers. Defaults to 5.
        """
        try:
            output = self.docker.compose.down(
                remove_orphans=remove_ophans,
                volumes=volumes,
                timeout=timeout,
                stream=True,
            )
            richprint.live_lines(output, padding=(0, 0, 0, 2))
        except DockerException as e:
            raise DockerComposeProjectFailedToRemoveError(
                self.compose_file_manager.compose_path, self.compose_file_manager.get_services_list()
            )

    def pull_images(self):
        """
        Pull docker images.
        """
        try:
            output = self.docker.compose.pull(stream=self.quiet)
            richprint.stdout.clear_live()
            if self.quiet:
                richprint.live_lines(output, padding=(0, 0, 0, 2))
        except DockerException as e:
            raise DockerComposeProjectFailedToPullImagesError(
                self.compose_file_manager.compose_path, self.compose_file_manager.get_services_list()
            )

    def logs(self, service: str, follow: bool = False):
        """
        Retrieve and print the logs for a specific service.

        Args:
            service (str): The name of the service.
            follow (bool, optional): Whether to continuously follow the logs. Defaults to False.
        """
        output = self.docker.compose.logs(services=[service], no_log_prefix=True, follow=follow, stream=True)
        for source, line in output:
            line = Text.from_ansi(line.decode())
            if source == "stdout":
                richprint.stdout.print(line)

    @property
    def running(self) -> bool:
        """
        Check if all services specified in the compose file are running.

        Returns:
            bool: True if all services are running, False otherwise.
        """
        services = self.compose_file_manager.get_services_list()
        running_status = self.get_services_running_status()

        if not running_status:
            return False

        for service in services:
            try:
                if not running_status[service] == "running":
                    return False
            except KeyError:
                return False
        return True

    def get_services_running_status(self) -> dict:
        """
        Get the running status of services in the Docker Compose file.

        Returns:
            A dictionary containing the running status of services.
            The keys are the service names, and the values are the container states.
        """
        services = self.compose_file_manager.get_services_list()
        containers = self.compose_file_manager.get_container_names().values()
        services_status = {}
        try:
            output = self.docker.compose.ps(service=services, format="json", all=True, stream=True)
            status: list = []
            for source, line in output:
                if source == "stdout":
                    current_status = json.loads(line.decode())
                    if type(current_status) == dict:
                        status.append(current_status)
                    else:
                        status += current_status

            # this is done to exclude docker runs using docker compose run command
            for container in status:
                if container["Name"] in containers:
                    services_status[container["Service"]] = container["State"]
            return services_status
        except DockerException as e:
            return {}

    def get_host_port_binds(self):
        """
        Get the list of published ports on the host for all containers.

        Returns:
            list: A list of published ports on the host.
        """
        try:
            output = self.docker.compose.ps(all=True, format="json", stream=True)
            status: list = []
            for source, line in output:
                if source == "stdout":
                    current_status = json.loads(line.decode())
                    if type(current_status) == dict:
                        status.append(current_status)
                    else:
                        status += current_status

            ports_info = []
            for container in status:
                try:
                    port_info = container["Publishers"]
                    if port_info:
                        ports_info = ports_info + port_info
                except KeyError as e:
                    pass

            published_ports = set()
            for port in ports_info:
                try:
                    published_port = port["PublishedPort"]
                    if published_port > 0:
                        published_ports.add(published_port)
                except KeyError as e:
                    pass
            return list(published_ports)

        except DockerException as e:
            return []

    def is_service_running(self, service):
        running_status = self.get_services_running_status()
        try:
            if running_status[service] == "running":
                return True
            else:
                return False
        except KeyError:
            return False

    def restart_service(self, services: List[str] = []):
        try:
            output = self.docker.compose.restart(services=services, stream=self.quiet)
            if self.quiet:
                richprint.live_lines(output, padding=(0, 0, 0, 2))
        except DockerException as e:
            raise DockerComposeProjectFailedToRestartError(
                self.compose_file_manager.compose_path, self.compose_file_manager.get_services_list()
            )



from pathlib import Path
from enum import Enum
from typing import Optional
from typer.core import TyperCommand
from frappe_manager.utils.cli_examples import get_examples_from_toml
import typer.rich_utils as ut
from rich.panel import Panel

# patch rich_format_help to display examples Panel
# save the function so that recurssion doesn't occur
rich_format_help_original = ut.rich_format_help

def print_fm_examples(*, obj, ctx, markup_mode):
    # utilising the original saved function
    rich_format_help_original(obj=obj, ctx=ctx, markup_mode=markup_mode)

    commands_stack = ctx.command_path.split(' ')[1:]

    new_doc = get_examples_from_toml(commands_stack=commands_stack, frappe_version=STABLE_APP_BRANCH_MAPPING_LIST["frappe"])

    if new_doc:
        import rich
        rich.print(
            Panel(
                new_doc,
                padding=ut.STYLE_OPTIONS_TABLE_PADDING,
                border_style=ut.STYLE_OPTIONS_PANEL_BORDER,
                title="Examples",
                title_align=ut.ALIGN_OPTIONS_PANEL,
            )
        )

ut.rich_format_help = print_fm_examples

# TODO configure this using config
# sites_dir = Path().home() / __name__.split(".")[0]
CLI_DIR = Path.home() / "frappe"
CLI_FM_CONFIG_PATH = CLI_DIR / "fm_config.toml"
CLI_SITES_ARCHIVE = CLI_DIR / "archived"
CLI_LOG_DIRECTORY = CLI_DIR / "logs"
CLI_BENCHES_DIRECTORY = CLI_DIR / "sites"
CLI_SERVICES_DIRECTORY = CLI_DIR / "services"

CLI_SERVICES_NGINX_PROXY_DIR = CLI_SERVICES_DIRECTORY / "nginx-proxy"
CLI_SERVICES_NGINX_PROXY_SSL_DIR = CLI_SERVICES_NGINX_PROXY_DIR / "ssl"

CLI_BENCH_CONFIG_FILE_NAME = "bench_config.toml"
SSL_RENEW_BEFORE_DAYS = 30


DEFAULT_EXTENSIONS = [
    "dbaeumer.vscode-eslint",
    "esbenp.prettier-vscode",
    "ms-python.python",
    "ms-python.debugpy",
    "ms-python.flake8",
    "ms-python.black-formatter",
    "visualstudioexptteam.vscodeintellicode",
    "VisualStudioExptTeam.intellicode-api-usage-examples",
]


class SiteServicesEnum(str, Enum):
    frappe = "frappe"
    nginx = "nginx"
    mariadb = "mariadb"
    redis_queue = "redis-queue"
    redis_cache = "redis-cache"
    redis_socketio = "redis-socketio"
    schedule = "schedule"
    socketio = "socketio"


STABLE_APP_BRANCH_MAPPING_LIST = {
    "frappe": "version-15",
    "erpnext": "version-15",
    "hrms": "version-15",
}


class EnableDisableOptionsEnum(str, Enum):
    enable = "enable"
    disable = "disable"



from pathlib import Path
from typing import List


class DockerComposeProjectFailedToStartError(Exception):
    def __init__(self, compose_path: Path, services: List[str], message='Failed to start compose services {}.') -> None:
        self.compose_path = compose_path
        self.services = services
        self.message = message.format(self.services)
        super().__init__(self.message)


class DockerComposeProjectFailedToStopError(Exception):
    def __init__(self, compose_path: Path, services: List[str], message='Failed to stop compose services {}.') -> None:
        self.compose_path = compose_path
        self.services = services
        self.message = message.format(self.services)
        super().__init__(self.message)


class DockerComposeProjectFailedToRemoveError(Exception):
    def __init__(
        self, compose_path: Path, services: List[str], message='Failed to remove compose services {}.'
    ) -> None:
        self.compose_path = compose_path
        self.services = services
        self.message = message.format(self.services)
        super().__init__(self.message)


class DockerComposeProjectFailedToPullImagesError(Exception):
    def __init__(
        self, compose_path: Path, services: List[str], message='Failed to pull compose services {} images.'
    ) -> None:
        self.compose_path = compose_path
        self.services = services
        self.message = message.format(self.services)
        super().__init__(self.message)


class DockerComposeProjectFailedToRestartError(Exception):
    def __init__(
        self, compose_path: Path, services: List[str], message='Failed to pull compose services {} images.'
    ) -> None:
        self.compose_path = compose_path
        self.services = services
        self.message = message.format(self.services)
        super().__init__(self.message)



import importlib
from cryptography.hazmat.backends import default_backend
from datetime import datetime
from cryptography import x509
from io import StringIO
import sys
from typing import Optional
from frappe_manager.utils.docker import run_command_with_exit_code
import requests
import subprocess
import platform
import time
import secrets
import grp
from pathlib import Path
import importlib.resources as pkg_resources
from rich.console import Console
from rich.traceback import Traceback
from frappe_manager.utils.site import is_fqdn
from frappe_manager.logger import log
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.site_manager import PREBAKED_SITE_APPS
from frappe_manager import CLI_BENCHES_DIRECTORY


def remove_zombie_subprocess_process(process):
    """
    This function iterates over a list of process IDs and terminates each process.

    Args:
        process (list): A list of process IDs to be terminated.

    Returns:
        None
    """
    if process:
        logger = log.get_logger()
        logger.cleanup("-" * 20)
        logger.cleanup(f"PROCESS: USED PROCESS {process}")

        import psutil

        for pid in process:
            try:
                process = psutil.Process(pid)
                process.terminate()
                logger.cleanup(f"Terminated Process {process.cmdline}:{pid}")
            except psutil.NoSuchProcess:
                logger.cleanup(f"{pid} Process not found")
            except psutil.AccessDenied:
                logger.cleanup(f"{pid} Permission denied")
        logger.cleanup("-" * 20)


def is_port_in_use(port):
    """
    Check if a port is in use or not.

    Args:
        port (int): The port number to check.

    Returns:
        bool: True if the port is in use, False otherwise.
    """
    import psutil

    for conn in psutil.net_connections():
        if conn.laddr.port == port and conn.status == "LISTEN":
            return True
    return False


def check_ports(ports):
    """
    Checks if the ports are in use.

    Args:
        ports (list): List of ports to be checked.

    Returns:
        list: List of binded ports (can be empty).
    """
    # TODO handle if ports are open using docker
    current_system = platform.system()
    already_binded = []
    for port in ports:
        if current_system == "Darwin":
            # Mac Os
            # check port using lsof command
            cmd = f"lsof -iTCP:{port} -sTCP:LISTEN -P -n"
            try:
                output = subprocess.run(cmd, check=True, shell=True, capture_output=True)
                if output.returncode == 0:
                    already_binded.append(port)
            except subprocess.CalledProcessError as e:
                pass
        else:
            # Linux or any other machines
            if is_port_in_use(port):
                already_binded.append(port)

    return already_binded


def check_and_display_port_status(ports_to_check: list, exclude=[]):
    """
    Check if the specified ports are already binded and display a message if they are.

    Args:
        ports_to_check (list): List of ports to check.
        exclude (list, optional): List of ports to exclude from checking. Defaults to [].
    """
    if exclude:
        # Removing elements present in remove_array from original_array
        ports_to_check = [x for x in exclude if x not in ports_to_check]

    if ports_to_check:
        already_binded = check_ports(ports_to_check)
        if already_binded:
            richprint.exit(
                f"Ports {', '.join(map(str, already_binded))} {'are' if len(already_binded) > 1 else 'is'} currently in use. Please free up these ports."
            )

def generate_random_text(length=50):
    """
    Generate a random text of specified length.

    Parameters:
    length (int): The length of the random text to be generated. Default is 50.

    Returns:
    str: The randomly generated text.
    """
    import random, string

    alphanumeric_chars = string.ascii_letters + string.digits
    return "".join(random.choice(alphanumeric_chars) for _ in range(length))


def is_cli_help_called(ctx):
    """
    Checks if the help is called for the CLI command.

    Args:
        ctx (object): The context object representing the CLI command.

    Returns:
        bool: True if the help command is called, False otherwise.
    """
    help_called = False
    # --help check

    if '--help' in " ".join(sys.argv[1:]):
        # is called command is sub command group
        return True

    try:
        for subtyper_command in ctx.command.commands[ctx.invoked_subcommand].commands.keys():
            check_command = " ".join(sys.argv[2:])
            if check_command == subtyper_command:
                if ctx.command.commands[ctx.invoked_subcommand].commands[subtyper_command].params:
                    help_called = True

    except AttributeError:
        help_called = False

    if not help_called:
        # is called command is sub command
        check_command = " ".join(sys.argv[1:])

        if check_command == ctx.invoked_subcommand:
            # is called command supports arguments then help called
            if ctx.command.commands[ctx.invoked_subcommand].params:
                help_called = True

            if not ctx.command.commands[ctx.invoked_subcommand].no_args_is_help:
                help_called = False

    return help_called


def get_current_fm_version():
    """
    Get the current version of the frappe-manager package.

    Returns:
        str: The current version of the frappe-manager package.
    """
    return importlib.metadata.version("frappe-manager")


def check_repo_exists(app_url: str, branch_name: str | None = None, exclude_dict: dict[str, str] = PREBAKED_SITE_APPS):
    """
    Check if a Frappe app exists on GitHub.

    Args:
        appname (str): The name of the Frappe app.
        branchname (str | None, optional): The name of the branch to check. Defaults to None.

    Returns:
        dict: A dictionary containing the existence status of the app and branch (if provided).
    """
    try:
        if app_url in exclude_dict:
            app = 200
        else:
            app = requests.get(app_url).status_code

        if branch_name:
            if branch_name in exclude_dict.values():
                branch = 200
            else:
                branch_url = f"{app_url}/tree/{branch_name}"
                branch = requests.get(branch_url).status_code

            return {
                "app": True if app == 200 else False,
                "branch": True if branch == 200 else False,
            }
        return {"app": True if app == 200 else False}

    except Exception as e:
        richprint.error(f"Not able to validate app {app_url} for branch [blue]{branch_name}[/blue]", e)


def check_frappe_app_exists(app: str, branch_name: Optional[str] = None):
    if "github.com" not in app:
        app = f"https://github.com/frappe/{app}"

    return check_repo_exists(app_url=app, branch_name=branch_name)


def represent_null_empty(string_null):
    """
    Replaces the string "null" with an empty string.

    Args:
        string_null (str): The input string.

    Returns:
        str: The modified string with "null" replaced by an empty string.
    """
    return string_null.replace("null", "")


def log_file(file, refresh_time: float = 0.1, follow: bool = False):
    """
    Generator function that yields new lines in a file

    Parameters:
    - file: The file object to read from
    - refresh_time: The time interval (in seconds) to wait before checking for new lines in the file (default: 0.1)
    - follow: If True, the function will continue to yield new lines as they are added to the file (default: False)

    Returns:
    - A generator that yields each new line in the file
    """
    file.seek(0)

    # start infinite loop
    while True:
        # read last line of file
        line = file.readline()
        if not line:
            if not follow:
                break
            # sleep if file hasn't been updated
            time.sleep(refresh_time)
            continue
        line = line.strip("\n")
        yield line


def get_container_name_prefix(site_name):
    """
    Returns the container name prefix by removing dots from the site name.

    Args:
        site_name (str): The name of the site.

    Returns:
        str: The container name prefix.
    """
    return site_name.replace(".", "")


def random_password_generate(password_length=13, symbols=False):
    # Define the character set to include symbols
    # symbols = "!@#$%^&*()_-+=[]{}|;:,.<>?`~"
    symbols = "!@%_-+?"

    # Generate a password without symbols using token_urlsafe

    generated_password = secrets.token_urlsafe(password_length)

    # Replace some characters with symbols in the generated password
    if symbols:
        password = "".join(c if secrets.choice([True, False]) else secrets.choice(symbols) for c in generated_password)
        return password

    return generated_password


# Retrieve Unix groups and their corresponding integer mappings
def get_unix_groups():
    groups = {}
    for group_entry in grp.getgrall():
        group_name = group_entry.gr_name
        groups[group_name] = group_entry.gr_gid
    return groups


def install_package(package_name, version):
    output = run_command_with_exit_code(
        [sys.executable, "-m", "pip", "install", f"{package_name}=={version}"], stream=True
    )
    richprint.live_lines(output)


def get_sitename_from_current_path() -> Optional[str]:
    current_path = Path().absolute()
    sites_path = CLI_BENCHES_DIRECTORY.absolute()

    if not current_path.is_relative_to(sites_path):
        return None

    sitename_list = list(current_path.relative_to(sites_path).parts)

    if not sitename_list:
        return None

    sitename = sitename_list[0]
    if is_fqdn(sitename):
        return sitename


def create_class_from_dict(class_name, attributes_dict):
    """
    Dynamically creates a class with properties based on the provided attributes dictionary.

    Parameters:
    class_name (str): The name of the class to be created.
    attributes_dict (dict): A dictionary where keys are the names of the properties and values are their default values.

    Returns:
    A new class with the specified properties and their default values.
    """
    return type(class_name, (object,), attributes_dict)


def create_symlink(source: Path, dest: Path):
    """
    Create a symbolic link pointing from dest to source.

    Parameters:
    - source (str): The source path that the symlink will point to.
    - dest (str): The destination path where the symlink will be created.

    Note: The function will overwrite the destination if a symlink already exists there.
    """

    # Convert the source and destination to Path objects

    # Remove the destination symlink/file/directory if it already exists
    if dest.exists() or dest.is_symlink():
        dest.unlink()

    # Create a symlink at the destination pointing to the source
    dest.symlink_to(source)


def get_template_path(file_name: str, template_dir: str = 'templates') -> Path:
    """
    Get the file path of a template.

    Args:
        file_name (str): The name of the template file.
        template_directory (str, optional): The directory where the templates are located. Defaults to "templates".

    Returns:
        Optional[str]: The file path of the template, or None if the template is not found.
    """
    template_path: str = f"{template_dir}/{file_name}"
    return get_frappe_manager_own_files(template_path)


def get_frappe_manager_own_files(file_path: str):
    return Path(str(pkg_resources.files("frappe_manager").joinpath(file_path)))


def rich_object_to_string(obj) -> str:
    """Convert a rich Traceback object to a string."""

    # Initialize a 'fake' console with StringIO to capture output
    capture_buffer = StringIO()

    fake_console = Console(force_terminal=False, file=capture_buffer)
    fake_console.print(obj, crop=False, overflow='ignore')

    captured_str = capture_buffer.getvalue()  # Retrieve the captured output as a string
    capture_buffer.close()
    return captured_str


def capture_and_format_exception(traceback_max_frames: int = 100) -> str:
    """Capture the current exception and return a formatted traceback string."""

    exc_type, exc_value, exc_traceback = sys.exc_info()  # Capture current exception info
    # Create a Traceback object with rich formatting
    #
    traceback = Traceback.from_exception(
        exc_type, exc_value, exc_traceback, show_locals=True, max_frames=traceback_max_frames
    )

    # Convert the Traceback object to a formatted string
    formatted_traceback = rich_object_to_string(traceback)

    return formatted_traceback


def pluralise(singular, count):
    return '{} {}{}'.format(count, singular, '' if count == 1 else 's')


def format_ssl_certificate_time_remaining(expiry_date: datetime):
    today_date = datetime.now(expiry_date.tzinfo)
    time_remaining = expiry_date - today_date
    day_count = time_remaining.days
    seconds_per_minute = 60
    seconds_per_hour = seconds_per_minute * 60
    seconds_unaccounted_for = time_remaining.seconds

    hours = int(seconds_unaccounted_for / seconds_per_hour)
    seconds_unaccounted_for -= hours * seconds_per_hour

    minutes = int(seconds_unaccounted_for / seconds_per_minute)

    return '{} {} {}'.format(pluralise('day', day_count), pluralise('hour', hours), pluralise('min', minutes))


def get_certificate_expiry_date(fullchain_path: Path) -> datetime:
    cert_content = fullchain_path.read_bytes()
    cert = x509.load_pem_x509_certificate(cert_content, default_backend())
    if hasattr(cert, 'not_valid_after_utc'):
        expiry_date: datetime = cert.not_valid_after_utc
    else:
        expiry_date: datetime = cert.not_valid_after
    return expiry_date



from typing import Optional
from pathlib import Path
from pydantic import BaseModel, EmailStr, Field
import tomlkit
from frappe_manager.migration_manager.version import Version
from frappe_manager import CLI_FM_CONFIG_PATH
from frappe_manager.utils.helpers import get_current_fm_version


class FMLetsencryptConfig(BaseModel):
    email: Optional[EmailStr] = Field(None, description="Email used by certbot.")
    api_token: Optional[str] = Field(None, description="Cloudflare API token used by Certbot.")
    api_key: Optional[str] = Field(None, description="Cloudflare Global API Key used by Certbot.")

    @property
    def exists(self):
        if self.api_token or self.api_key:
            return True

        return False

    def get_toml_doc(self):
        model_dict = self.model_dump(exclude_none=True)
        toml_doc = tomlkit.document()

        for key, value in model_dict.items():
            if isinstance(value, Path):
                toml_doc[key] = str(value.absolute())
            else:
                toml_doc[key] = value
        return toml_doc

    @classmethod
    def import_from_toml_doc(cls, toml_doc):
        config_object = cls(**toml_doc)
        return config_object


class FMConfigManager(BaseModel):
    root_path: Path
    version: Version
    letsencrypt: FMLetsencryptConfig = Field(default=FMLetsencryptConfig())

    def export_to_toml(self, path: Path = CLI_FM_CONFIG_PATH) -> bool:
        exclude = {'root_path'}

        if not self.letsencrypt.email and not self.letsencrypt.api_key and not self.letsencrypt.api_token:
            exclude.add('letsencrypt')
        else:
            if self.letsencrypt.email == 'dummy@fm.fm':
                exclude.add('letsencrypt')

        if self.version < Version('0.13.0'):
            path = CLI_FM_CONFIG_PATH.parent / '.fm.toml'

        # Convert the BenchConfig instance to a dictionary
        fm_config_dict = self.model_dump(exclude=exclude, exclude_none=True)

        # transform structure if needed
        fm_config_dict['version'] = self.version.version

        toml_doc = tomlkit.document()

        for key, value in fm_config_dict.items():
            toml_doc[key] = value

        try:
            with open(path, 'w') as f:
                f.write(tomlkit.dumps(toml_doc))
            return True
        except Exception as e:
            return False

    @classmethod
    def import_from_toml(cls, path: Path = CLI_FM_CONFIG_PATH) -> "FMConfigManager":
        # previously the name of the fm config was .fm.toml now changed to fm_cofig.toml
        input_data = {}

        old_config_path = path.parent / '.fm.toml'

        input_data['version'] = Version('0.8.3')
        input_data['letsencrypt'] = FMLetsencryptConfig(email=None, api_key=None, api_token=None)
        input_data['root_path'] = str(path)

        if old_config_path.exists():
            old_data = tomlkit.parse(old_config_path.read_text())
            input_data['version'] = Version(old_data.get('version', '0.8.3'))
        elif path.exists():
            data = tomlkit.parse(path.read_text())
            input_data['version'] = Version(data.get('version', get_current_fm_version()))
            input_data['letsencrypt'] = FMLetsencryptConfig(
                **data.get('letsencrypt', {'email': None, 'api_key': None, 'api_token': None})
            )

        fm_config_instance = cls(**input_data)
        return fm_config_instance



import json
from pathlib import Path
import platform
from typing import Dict
from frappe_manager import CLI_SERVICES_DIRECTORY
from frappe_manager.compose_manager.ComposeFile import ComposeFile

from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.utils.site import get_bench_db_connection_info


class MigrationBench:
    def __init__(self, name: str, path: Path) -> None:
        self.name = name
        self.path = path
        self.compose_project = ComposeProject(
            ComposeFile(
                self.path / 'docker-compose.yml',
                'docker-compose.tmpl',
                template_dir='migration_manager/base_templates',
            )
        )
        self.workers_compose_project = ComposeProject(
            ComposeFile(self.path / 'docker-compose.workers.yml', 'docker-compose.workers.tmpl')
        )

    def get_db_connection_info(self):
        return get_bench_db_connection_info(self.name, self.path)

    def common_bench_config_set(self, config: dict):
        """
        Sets the values in the common_site_config.json file.
        Args:
            config (dict): A dictionary containing the key-value pairs to be set in the common_site_config.json file.
        Returns:
            bool: True if the values are successfully set, False otherwise.
        """
        common_bench_config_path = self.path / "workspace/frappe-bench/sites/common_site_config.json"

        if not common_bench_config_path.exists():
            return False

        common_site_config = {}

        with open(common_bench_config_path, "r") as f:
            common_site_config = json.load(f)

        try:
            for key, value in config.items():
                common_site_config[key] = value
            with open(common_bench_config_path, "w") as f:
                json.dump(common_site_config, f)
            return True
        except KeyError as e:
            return False


class MigrationBenches:
    def __init__(self, benches_path: Path) -> None:
        self.benches_path = benches_path

    def get_all_benches(self, exclude=[]):
        # get list of all sites
        benches: Dict[str, Path] = {}
        for dir in self.benches_path.iterdir():
            if dir.is_dir() and dir.parts[-1] not in exclude:
                name = dir.parts[-1]
                dir = dir / "docker-compose.yml"
                if dir.exists():
                    benches[name] = dir
        return benches

    def stop_benches(self, timeout: int = 100):
        compose_list = self.get_all_benches()
        for name, compose_path in compose_list.items():
            bench = MigrationBench(name, compose_path.parent)
            bench.compose_project.stop_service(timeout=timeout)


class MigrationServicesManager:
    def __init__(self, services_path: Path = CLI_SERVICES_DIRECTORY):
        self.services_path = services_path

        template_name = 'docker-compose.services.tmpl'

        if platform.system() == "Darwin":
            template_name = 'docker-compose.services.osx.tmpl'

        self.compose_project = ComposeProject(
            ComposeFile(
                self.services_path / 'docker-compose.yml',
                template_name,
                template_dir='migration_manager/base_templates',
            )
        )



import json
import os
import copy
from pathlib import Path
import tomlkit
from frappe_manager.compose_manager import DockerVolumeMount
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from frappe_manager.migration_manager.migration_base import MigrationBase
from frappe_manager.migration_manager.migration_exections import MigrationExceptionInBench
from frappe_manager.migration_manager.migration_helpers import (
    MigrationBench,
    MigrationBenches,
    MigrationServicesManager,
)
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.migration_manager.version import Version
from frappe_manager import CLI_DIR, CLI_SERVICES_DIRECTORY
from frappe_manager.utils.helpers import get_container_name_prefix
from frappe_manager.docker_wrapper.DockerClient import DockerClient
from frappe_manager.migration_manager.backup_manager import BackupManager


class MigrationV0130(MigrationBase):
    version = Version("0.13.0")

    def init(self):
        self.cli_dir: Path = Path.home() / 'frappe'
        self.benches_dir = self.cli_dir / "sites"
        self.backup_manager = BackupManager(name=str(self.version), benches_dir=self.benches_dir)
        self.benches_manager = MigrationBenches(self.benches_dir)
        self.services_manager: MigrationServicesManager = MigrationServicesManager(
            services_path=self.cli_dir / 'services'
        )

    def migrate_services(self):
        # remove version from services yml
        try:
            del self.services_manager.compose_project.compose_file_manager.yml['version']
        except KeyError:
            self.logger.warning("[services]: 'version' attribute not found in compose file.")
            pass

        # include new volume info
        services_volume_list = self.services_manager.compose_project.compose_file_manager.get_service_volumes(
            'global-nginx-proxy'
        )
        host_dir = CLI_SERVICES_DIRECTORY / 'nginx-proxy' / 'ssl'

        host_dir.mkdir(exist_ok=True, parents=True)

        container_dir = '/usr/share/nginx/ssl'

        ssl_volume = DockerVolumeMount(
            host=host_dir,
            container=container_dir,
            type='bind',
            compose_path=self.services_manager.compose_project.compose_file_manager.compose_path,
        )
        services_volume_list.append(ssl_volume)

        self.services_manager.compose_project.compose_file_manager.set_service_volumes(
            'global-nginx-proxy', volumes=services_volume_list
        )
        self.services_manager.compose_project.compose_file_manager.set_version(self.version.version_string())
        self.services_manager.compose_project.compose_file_manager.write_to_file()

        if self.services_manager.compose_project.is_service_running('global-nginx-proxy'):
            self.services_manager.compose_project.docker.compose.up(services=['global-nginx-proxy'])

        # rename main config
        fm_config_path = CLI_DIR / 'fm_config.toml'
        old_fm_config_path = CLI_DIR / '.fm.toml'

        if old_fm_config_path.exists():
            old_fm_config_path.rename(fm_config_path)

    def migrate_bench(self, bench: MigrationBench):
        bench_common_site_config = bench.path / "workspace" / "frappe-bench" / "sites" / "common_site_config.json"
        common_site_config_json = json.loads(bench_common_site_config.read_bytes())

        if 'mail_server' in common_site_config_json:
            common_site_config_json['mail_server'] = get_container_name_prefix(bench.name)
            bench_common_site_config.write_text(json.dumps(common_site_config_json))

        bench.compose_project.down_service(volumes=False)
        self.migrate_bench_compose(bench)

    def undo_bench_migrate(self, bench: MigrationBench):
        richprint.change_head("Removing Admin Tools compose file")

        admin_tools_compose_path = bench.path / 'docker-compose.admin-tools.yml'

        if admin_tools_compose_path.exists():
            admin_tools_compose_path.unlink()

        richprint.print("Removed Admin Tools compose file")

    def migrate_bench_compose(self, bench: MigrationBench):
        richprint.change_head("Migrating bench compose")

        if not bench.compose_project.compose_file_manager.exists():
            richprint.print(f"Failed to migrate {bench.name} compose file.")
            raise MigrationExceptionInBench(f"{bench.compose_project.compose_file_manager.compose_path} not found.")

        # get all the payloads
        envs = bench.compose_project.compose_file_manager.get_all_envs()

        # add HSTS=off envionment variable for all the benches
        envs["nginx"]["HSTS"] = 'off'

        if 'ENABLE_SSL' in envs['nginx']:
            try:
                del envs['nginx']['ENABLE_SSL']
            except KeyError:
                self.logger.warning(f"{bench.name} 'ENABLE_SSL' nginx's env not found.")
                pass

        # create new html in configs/nginx/html compose directory
        html_nginx_configs_path = bench.path / 'configs' / 'nginx' / 'html'
        html_nginx_configs_path.mkdir(parents=True, exist_ok=True)

        bench_config_path = bench.path / 'bench_config.toml'

        # create bench config
        frappe = envs.get('frappe', {})

        apps_list = frappe.get('APPS_LIST', None)

        if apps_list:
            apps_list = apps_list.split(',')
        else:
            apps_list = []

        developer_mode = frappe.get('DEVELOPER_MODE', True)
        name = frappe.get('SITENAME', bench.name)

        bench_config = tomlkit.document()

        bench_config['name'] = name
        bench_config['developer_mode'] = developer_mode
        bench_config['admin_tools'] = True
        bench_config['environment_type'] = 'dev'

        with open(bench_config_path, 'w') as f:
            f.write(tomlkit.dumps(bench_config))

        images_info = bench.compose_project.compose_file_manager.get_all_images()

        # change image frappe, socketio, schedule
        self.frappe_image_info = images_info['frappe']
        nginx_image_info = images_info['nginx']
        nginx_image_info['tag'] = self.version.version_string()
        self.frappe_image_info['tag'] = self.version.version_string()

        # setting all images
        images_info['frappe'] = self.frappe_image_info
        images_info['socketio'] = self.frappe_image_info
        images_info['schedule'] = self.frappe_image_info
        images_info['nginx'] = nginx_image_info

        for image in [self.frappe_image_info, nginx_image_info]:
            pull_image = f"{image['name']}:{image['tag']}"
            richprint.change_head(f"Pulling Image {pull_image}")
            output = DockerClient().pull(container_name=pull_image, stream=True)
            richprint.live_lines(output, padding=(0, 0, 0, 2))
            richprint.print(f"Image pulled [blue]{pull_image}[/blue]")

        self.migrate_admin_tools_compose(bench)

        # remove adminer and mailhog from main bench compose
        del bench.compose_project.compose_file_manager.yml['services']['adminer']

        del bench.compose_project.compose_file_manager.yml['services']['mailhog']

        # remove version from compose
        try:
            del bench.compose_project.compose_file_manager.yml['version']
        except KeyError:
            self.logger.warning(f"{bench.name} 'version' attribute not found in compose file.")
            pass

        # include new volume info
        bench_volume_list = bench.compose_project.compose_file_manager.get_service_volumes('nginx')
        container_dir = '/usr/share/nginx/html'
        ssl_volume = DockerVolumeMount(
            host=html_nginx_configs_path,
            container=container_dir,
            type='bind',
            compose_path=bench.compose_project.compose_file_manager.compose_path,
        )
        bench_volume_list.append(ssl_volume)

        bench.compose_project.compose_file_manager.set_service_volumes('nginx', volumes=bench_volume_list)

        bench.compose_project.compose_file_manager.set_all_images(images_info)

        bench.compose_project.compose_file_manager.set_all_envs(envs)
        bench.compose_project.compose_file_manager.set_version(str(self.version))
        bench.compose_project.compose_file_manager.write_to_file()

        richprint.print(f"Migrated [blue]{bench.name}[/blue] compose file.")
        self.migrate_workers_compose(bench)

    def migrate_workers_compose(self, bench: MigrationBench):
        if bench.workers_compose_project.compose_file_manager.compose_path.exists():
            richprint.print("Migrating workers compose")

            workers_image_info = bench.workers_compose_project.compose_file_manager.get_all_images()
            for worker in workers_image_info.keys():
                workers_image_info[worker] = self.frappe_image_info

            try:
                del bench.workers_compose_project.compose_file_manager.yml['version']
            except KeyError:
                self.logger.warning(f"{bench.name} workers 'version' attribute not found in compose file.")
                pass

            bench.workers_compose_project.compose_file_manager.set_top_networks_name(
                "site-network", get_container_name_prefix(bench.name)
            )
            bench.workers_compose_project.compose_file_manager.set_container_names(
                get_container_name_prefix(bench.name)
            )
            bench.compose_project.compose_file_manager.set_version(str(self.version))
            bench.workers_compose_project.compose_file_manager.write_to_file()

            richprint.print(f"Migrated [blue]{bench.name}[/blue] workers compose file.")

    def migrate_admin_tools_compose(self, bench: MigrationBench):
        richprint.change_head("Create Admin Tools")

        bench_compose_yml = copy.deepcopy(bench.compose_project.compose_file_manager.yml)

        adminer_compose_service_config = bench_compose_yml['services']['adminer']
        mailhog_compose_service_config = bench_compose_yml['services']['mailhog']

        network_compose_config = bench_compose_yml.get('networks', {})

        if 'global-frontend-network' in network_compose_config:
            try:
                del network_compose_config['global-frontend-network']
            except KeyError:
                pass

        admin_tools_compose = {}
        admin_tools_compose['services'] = {
            "mailhog": mailhog_compose_service_config,
            "adminer": adminer_compose_service_config,
        }

        admin_tools_compose['networks'] = network_compose_config

        # this should run before bench compose migrate or in betwen since this depends on i
        admin_tools_compose_path = bench.path / 'docker-compose.admin-tools.yml'

        # handle admin tool migration
        admin_tools_compose_file_manager = ComposeFile(
            admin_tools_compose_path, template_name='docker-compose.admin-tools.tmpl'
        )

        admin_tools_compose_file_manager.yml = admin_tools_compose
        admin_tools_compose_file_manager.set_version(self.version.version_string())
        admin_tools_compose_file_manager.write_to_file()

        # create custom nginx directory for adming tool location config

        nginx_conf_dir = bench.path / 'configs' / 'nginx' / 'conf'
        if nginx_conf_dir.exists():
            nginx_custom_dir = nginx_conf_dir / 'custom'
            nginx_custom_dir.mkdir(parents=True, exist_ok=True)

        # change mailhog config in common_site_confiig.json
        common_bench_config_path = bench.path / "workspace/frappe-bench/sites/common_site_config.json"

        current_common_site_config = json.loads(common_bench_config_path.read_text())

        new_conf = {
            "mail_port": 1025,
            "mail_server": f"{get_container_name_prefix(bench.name)}-mailhog",
            "disable_mail_smtp_authentication": 1,
        }

        for key, value in new_conf.items():
            current_common_site_config[key] = value

        common_bench_config_path.write_text(json.dumps(current_common_site_config))

        richprint.change_head(f"Created {bench.name} Admin Tools.")



from pathlib import Path
from frappe_manager.migration_manager.migration_base import MigrationBase
from frappe_manager.docker_wrapper.DockerClient import DockerClient
from frappe_manager.migration_manager.migration_exections import MigrationExceptionInBench
from frappe_manager.migration_manager.migration_helpers import (
    MigrationBench,
    MigrationBenches,
    MigrationServicesManager,
)
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.migration_manager.version import Version
from frappe_manager.utils.helpers import get_container_name_prefix
from frappe_manager.migration_manager.backup_manager import BackupManager


class MigrationV0120(MigrationBase):
    version = Version("0.12.0")

    def init(self):
        self.cli_dir: Path = Path.home() / 'frappe'
        self.benches_dir = self.cli_dir / "sites"
        self.backup_manager = BackupManager(name=str(self.version), benches_dir=self.benches_dir)
        self.benches_manager = MigrationBenches(self.benches_dir)
        self.services_manager: MigrationServicesManager = MigrationServicesManager(
            services_path=self.cli_dir / 'services'
        )

    def migrate_services(self):
        # Pulling latest image
        self.image_info = {"tag": self.version.version_string(), "name": "ghcr.io/rtcamp/frappe-manager-frappe"}
        pull_image = f"{self.image_info['name']}:{self.image_info['tag']}"

        richprint.change_head(f"Pulling Image {pull_image}")
        output = DockerClient().pull(container_name=pull_image, stream=True)
        richprint.live_lines(output, padding=(0, 0, 0, 2))
        richprint.print(f"Image pulled [blue]{pull_image}[/blue]")

    def migrate_bench(self, bench: MigrationBench):
        bench.compose_project.down_service(volumes=False)
        self.migrate_bench_compose(bench)
        self.migrate_workers_compose(bench)

    def migrate_bench_compose(self, bench: MigrationBench):
        richprint.change_head("Migrating bench compose")

        if not bench.compose_project.compose_file_manager.exists():
            richprint.print(f"Failed to migrate {bench.name} compose file.")
            raise MigrationExceptionInBench(f"{bench.compose_project.compose_file_manager.compose_path} not found.")

        images_info = bench.compose_project.compose_file_manager.get_all_images()

        # for all services
        images_info["frappe"] = self.image_info
        images_info["socketio"] = self.image_info
        images_info["schedule"] = self.image_info

        compose_yml = bench.compose_project.compose_file_manager.yml
        # remove restart: from all the services
        for service in compose_yml["services"]:
            try:
                del compose_yml["services"][service]["restart"]
            except KeyError as e:
                self.logger.error(
                    f"{bench.name} worker not able to delete 'restart: always' attribute from compose file.{e}"
                )
                pass

        richprint.print("Removed 'restart: always'.")

        bench.compose_project.compose_file_manager.set_version(str(self.version))
        bench.compose_project.compose_file_manager.set_all_images(images_info)
        bench.compose_project.compose_file_manager.write_to_file()

        richprint.print(f"Migrated {bench.name} compose file.")

    def migrate_workers_compose(self, bench: MigrationBench):
        if bench.workers_compose_project.compose_file_manager.compose_path.exists():
            richprint.print("Migrating workers compose")

            # workers image set
            workers_info = bench.workers_compose_project.compose_file_manager.get_all_images()

            for worker in workers_info.keys():
                workers_info[worker] = self.image_info

            worker_compose_yml = bench.workers_compose_project.compose_file_manager.yml
            for service in worker_compose_yml["services"]:
                try:
                    del worker_compose_yml["services"][service]["restart"]
                except KeyError as e:
                    self.logger.error(
                        f"{bench.name} worker not able to delete 'restart: always' attribute from compose file.{e}"
                    )
                    pass

            bench.workers_compose_project.compose_file_manager.set_top_networks_name(
                "site-network", get_container_name_prefix(bench.name)
            )
            bench.workers_compose_project.compose_file_manager.set_all_images(workers_info)

            bench.workers_compose_project.compose_file_manager.set_container_names(
                get_container_name_prefix(bench.name)
            )
            bench.workers_compose_project.compose_file_manager.write_to_file()

        richprint.print(f"Migrated [blue]{bench.name}[/blue] compose file.")



from frappe_manager.utils.site import pull_docker_images
import typer
import requests
import json
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.utils.helpers import get_current_fm_version, install_package

update_app = typer.Typer(rich_markup_mode="rich")


@update_app.callback(invoke_without_command=True)
def update_callback(
    ctx: typer.Context,
):
    if ctx.invoked_subcommand == None:
        richprint.change_head("Checking for udpates")
        url = "https://pypi.org/pypi/frappe-manager/json"
        try:
            update_info = requests.get(url, timeout=2)
            update_info = json.loads(update_info.text)
            fm_version = get_current_fm_version()
            latest_version = update_info["info"]["version"]
            if not fm_version == latest_version:
                update_msg = (
                    f":arrows_counterclockwise: New update available [blue][bold]v{latest_version}[/bold][/blue]"
                    "\nDo you want to update ?"
                )
                continue_update = richprint.prompt_ask(prompt=update_msg, choices=["yes", "no"])

                if continue_update == 'yes':
                    install_package("frappe-manager", latest_version)
        except Exception as e:
            richprint.exit(f"Error occured while updating the app : {e}")


@update_app.command()
def images(
    ctx: typer.Context,
):
    """Pull latest FM stack docker images."""

    pull_docker_images()



from copy import deepcopy
from pathlib import Path
from typing import List, TYPE_CHECKING
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.site_manager.site_exceptions import BenchWorkersSupervisorConfigurtionNotFoundError
from frappe_manager.utils.helpers import get_container_name_prefix, get_current_fm_version

if TYPE_CHECKING:
    from frappe_manager.site_manager.site import Bench


class BenchWorkers:
    def __init__(self, bench: 'Bench', verbose: bool = True):
        self.bench = bench
        self.compose_path = self.bench.path / "docker-compose.workers.yml"
        self.config_dir = self.bench.path / "workspace" / "frappe-bench" / "config"
        self.supervisor_config_path = self.config_dir / "supervisor.conf"
        self.quiet = not verbose
        self.compose_project = ComposeProject(
            ComposeFile(self.compose_path, template_name='docker-compose.workers.tmpl')
        )

    def get_expected_workers(self) -> List[str]:
        richprint.change_head("Checking workers info.")

        workers_supervisor_conf_paths = []

        for file_path in self.config_dir.iterdir():
            file_path_abs = str(file_path.absolute())
            if file_path.is_file():
                if file_path_abs.endswith(".workers.fm.supervisor.conf"):
                    workers_supervisor_conf_paths.append(file_path)

        if len(workers_supervisor_conf_paths) == 0:
            raise BenchWorkersSupervisorConfigurtionNotFoundError(self.bench.name, self.config_dir)

        workers_expected_service_names = []

        for worker_name in workers_supervisor_conf_paths:
            worker_name = worker_name.name
            worker_name = worker_name.replace("frappe-bench-frappe-", "")
            worker_name = worker_name.replace(".workers.fm.supervisor.conf", "")
            workers_expected_service_names.append(worker_name)

        workers_expected_service_names.sort()

        return workers_expected_service_names

    def is_new_workers_added(self) -> bool:
        if not self.compose_project.compose_file_manager.is_template_loaded:
            prev_workers = self.compose_project.compose_file_manager.get_services_list()
            prev_workers.sort()
            expected_workers = self.get_expected_workers()
            # get custom workers from common_site_config.json
            # common_site_config_data = self.bench.get_common_bench_config()

            # if 'workers' in common_site_config_data:
            #     custom_workers: List[str] = common_site_config_data['workers'].keys()
            #     for worker in custom_workers:
            #         worker = f'{worker}-worker'
            #         if worker not in prev_workers:
            #             return False
            return prev_workers == expected_workers
        else:
            return False

    def generate_compose(self):
        if not self.compose_path.exists():
            richprint.print("Workers compose not present. Generating...")
        else:
            richprint.print("Workers configuration changed. Recreating compose...")

        # create compose file for workers
        self.compose_project.compose_file_manager.yml = self.compose_project.compose_file_manager.load_template()

        template_worker_config = self.compose_project.compose_file_manager.yml["services"]["worker-name"]

        del self.compose_project.compose_file_manager.yml["services"]["worker-name"]

        workers_expected_service_names = self.get_expected_workers()

        import os

        for worker in workers_expected_service_names:
            worker_config = deepcopy(template_worker_config)

            # setting environments
            worker_config["environment"]["WAIT_FOR"] = str(worker_config["environment"]["WAIT_FOR"]).replace(
                "{worker-name}", worker
            )
            worker_config["environment"]["COMMAND"] = str(worker_config["environment"]["COMMAND"]).replace(
                "{worker-name}", worker
            )
            worker_config["environment"]["USERID"] = os.getuid()
            worker_config["environment"]["USERGROUP"] = os.getgid()

            self.compose_project.compose_file_manager.yml["services"][worker] = worker_config

        self.compose_project.compose_file_manager.set_container_names(get_container_name_prefix(self.bench.name))

        self.compose_project.compose_file_manager.set_version(get_current_fm_version())

        # set network name
        self.compose_project.compose_file_manager.yml["networks"]["site-network"]["name"] = (
            self.bench.name.replace(".", "") + f"-network"
        )
        self.compose_project.compose_file_manager.write_to_file()



from pathlib import Path
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.docker_wrapper.DockerClient import DockerClient
from frappe_manager.migration_manager.migration_base import MigrationBase
from frappe_manager.migration_manager.migration_exections import MigrationExceptionInBench
from frappe_manager.migration_manager.migration_helpers import (
    MigrationBench,
    MigrationBenches,
    MigrationServicesManager,
)
from frappe_manager.migration_manager.version import Version
from frappe_manager.migration_manager.backup_manager import BackupManager
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.services_manager.database_service_manager import (
    DatabaseServerServiceInfo,
    DatabaseServiceManager,
    MariaDBManager,
)
from frappe_manager.utils.helpers import get_container_name_prefix


class MigrationV0150(MigrationBase):
    version = Version("0.15.0")

    def init(self):
        self.cli_dir: Path = Path.home() / 'frappe'
        self.benches_dir = self.cli_dir / "sites"
        self.backup_manager = BackupManager(name=str(self.version), benches_dir=self.benches_dir)
        self.benches_manager = MigrationBenches(self.benches_dir)
        self.services_manager: MigrationServicesManager = MigrationServicesManager(
            services_path=self.cli_dir / 'services'
        )
        self.pulled_images_list = []

    def migrate_services(self):
        images_info = self.services_manager.compose_project.compose_file_manager.get_all_images()
        images_info['global-nginx-proxy'] = {'name': 'jwilder/nginx-proxy', 'tag': '1.6'}

        # pulling nginx proxy image
        pull_image = f"{images_info['global-nginx-proxy']['name']}:{images_info['global-nginx-proxy']['tag']}"
        richprint.change_head(f"Pulling Image {pull_image}")
        output = DockerClient().pull(container_name=pull_image, stream=True)
        richprint.live_lines(output, padding=(0, 0, 0, 2))
        richprint.print(f"Image pulled [blue]{pull_image}[/blue]")

        richprint.change_head("Migrating services compose")

        # add ACME_HTTP_CHALLENGE_LOCATION: false env
        envs = self.services_manager.compose_project.compose_file_manager.get_all_envs()
        envs['global-nginx-proxy'] = {'ACME_HTTP_CHALLENGE_LOCATION': False}

        self.services_manager.compose_project.compose_file_manager.set_all_envs(envs)
        self.services_manager.compose_project.compose_file_manager.set_version(str(self.version))
        self.services_manager.compose_project.compose_file_manager.write_to_file()
        richprint.print("Migrated services compose")
        richprint.change_head("Restarting services")
        self.services_manager.compose_project.start_service(force_recreate=True)

        services_database_manager: DatabaseServiceManager = MariaDBManager(
            DatabaseServerServiceInfo.import_from_compose_file('global-db', self.services_manager.compose_project),
            self.services_manager.compose_project,
        )
        # wait till db starts
        services_database_manager.wait_till_db_start()

        richprint.print("Restarted services")

    def migrate_bench(self, bench: MigrationBench):
        bench.compose_project.down_service(volumes=True)
        richprint.change_head("Migrating bench compose")

        if not bench.compose_project.compose_file_manager.exists():
            richprint.print(f"Failed to migrate {bench.name} compose file.")
            raise MigrationExceptionInBench(f"{bench.compose_project.compose_file_manager.compose_path} not found.")

        images_info = bench.compose_project.compose_file_manager.get_all_images()

        frappe_image_info = images_info['frappe']
        frappe_image_info['tag'] = self.version.version_string()

        redis_image_info = images_info['redis-cache']
        redis_image_info['tag'] = '6.2-alpine'

        nginx_image_info = images_info['nginx']
        nginx_image_info['tag'] = self.version.version_string()

        # change image nginx
        images_info['nginx'] = nginx_image_info

        # change image frappe, socketio, schedule
        images_info['frappe'] = frappe_image_info
        images_info['socketio'] = frappe_image_info
        images_info['schedule'] = frappe_image_info

        # change image for redis
        images_info['redis-cache'] = redis_image_info
        images_info['redis-queue'] = redis_image_info
        images_info['redis-socketio'] = redis_image_info

        # remove default.conf file from nginx
        nginx_default_conf_path = bench.path / 'configs' / 'nginx' / 'conf' / 'conf.d' / 'default.conf'

        if nginx_default_conf_path.exists():
            nginx_default_conf_path.unlink()

        for image in [frappe_image_info, redis_image_info, nginx_image_info]:
            pull_image = f"{image['name']}:{image['tag']}"
            if pull_image not in self.pulled_images_list:
                richprint.change_head(f"Pulling Image {pull_image}")
                output = DockerClient().pull(container_name=pull_image, stream=True)
                richprint.live_lines(output, padding=(0, 0, 0, 2))
                richprint.print(f"Image pulled [blue]{pull_image}[/blue]")
                self.pulled_images_list.append(pull_image)

        bench.compose_project.compose_file_manager.set_all_images(images_info)
        bench.compose_project.compose_file_manager.set_version(str(self.version))
        bench.compose_project.compose_file_manager.write_to_file()

        self.migrate_workers_compose(bench)
        self.migrate_admin_tools_compose(bench)

    def migrate_workers_compose(self, bench: MigrationBench):
        if bench.workers_compose_project.compose_file_manager.compose_path.exists():
            richprint.change_head("Migrating workers compose")
            workers_image_info = bench.workers_compose_project.compose_file_manager.get_all_images()

            for worker in workers_image_info.keys():
                workers_image_info[worker]['tag'] = self.version.version_string()

            bench.workers_compose_project.compose_file_manager.set_top_networks_name(
                "site-network", get_container_name_prefix(bench.name)
            )
            bench.workers_compose_project.compose_file_manager.set_container_names(
                get_container_name_prefix(bench.name)
            )
            bench.workers_compose_project.compose_file_manager.set_all_images(workers_image_info)
            bench.workers_compose_project.compose_file_manager.set_version(str(self.version))
            bench.workers_compose_project.compose_file_manager.write_to_file()
            richprint.print(f"Migrated [blue]{bench.name}[/blue] workers compose file.")

    def migrate_admin_tools_compose(self, bench: MigrationBench):
        admin_tool_compose_file = bench.path / 'docker-compose.admin-tools.yml'
        if admin_tool_compose_file.exists():
            admin_tool_compose_file_manager = ComposeFile(admin_tool_compose_file, 'docker-compose.admin-tools.tmpl')
            admin_tool_compose_project = ComposeProject(admin_tool_compose_file_manager)

            richprint.change_head("Migrating admin-tools compose")

            admin_tools_image_info = admin_tool_compose_project.compose_file_manager.get_all_images()
            admin_tools_image_info['adminer']['tag'] = '4'

            for image in [admin_tools_image_info['adminer']]:
                pull_image = f"{image['name']}:{image['tag']}"

                if pull_image not in self.pulled_images_list:
                    richprint.change_head(f"Pulling Image {pull_image}")
                    output = DockerClient().pull(container_name=pull_image, stream=True)
                    richprint.live_lines(output, padding=(0, 0, 0, 2))
                    richprint.print(f"Image pulled [blue]{pull_image}[/blue]")
                    self.pulled_images_list.append(pull_image)

            admin_tool_compose_project.compose_file_manager.set_top_networks_name(
                "site-network", get_container_name_prefix(bench.name)
            )
            admin_tool_compose_project.compose_file_manager.set_all_images(admin_tools_image_info)
            admin_tool_compose_project.compose_file_manager.set_container_names(get_container_name_prefix(bench.name))
            admin_tool_compose_project.compose_file_manager.set_version(str(self.version))
            admin_tool_compose_project.compose_file_manager.write_to_file()

            richprint.print(f"Migrated [blue]{bench.name}[/blue] admin-tools compose file.")



from subprocess import Popen, run, TimeoutExpired, CalledProcessError
from pathlib import Path
from typing import Iterable, List, Union, Literal, Optional, Tuple

import shlex

from frappe_manager.utils.docker import (
    SubprocessOutput,
    parameters_to_options,
    run_command_with_exit_code,
)


# Docker Compose version 2.18.1
class DockerComposeWrapper:
    """
    This class provides one to one mapping between docker compose cli each function.
    Only this args have are different use case.

    Args:
        stream (bool, optional): A boolean flag indicating whether to stream the output of the command as it runs.
            If set to True, the output will be displayed in real-time. If set to False, the output will be
            displayed after the command completes. Defaults to False.
    """

    def __init__(self, path: Path, timeout: int = 100):
        # requires valid path directory
        # directory where docker-compose resides
        self.compose_file_path = path.absolute()

        self.docker_compose_cmd = [
            "docker",
            "compose",
            "-f",
            self.compose_file_path.as_posix(),
        ]

    def up(
        self,
        services: list[str] = [],
        detach: bool = True,
        build: bool = False,
        remove_orphans: bool = False,
        no_recreate: bool = False,
        force_recreate: bool = False,
        always_recreate_deps: bool = False,
        quiet_pull: bool = False,
        pull: Literal["missing", "never", "always"] = "missing",
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()

        remove_parameters = ["services", "stream"]

        up_cmd: list = ["up"]
        up_cmd += services

        up_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        # subprocess_env = dict(os.environ)
        # subprocess_env.update(env)

        iterator = run_command_with_exit_code(self.docker_compose_cmd + up_cmd, stream=stream)
        return iterator

    def down(
        self,
        timeout: int = 100,
        remove_orphans: bool = False,
        rmi: Union[bool, Literal["all", "local"]] = False,
        volumes: bool = False,
        dry_run: bool = False,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()
        parameters["timeout"] = str(timeout)

        down_cmd: list[str] = ["down"]

        remove_parameters = ["stream"]

        if not rmi:
            remove_parameters.append("rmi")
        else:
            parameters["rmi"] = "all"

        down_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        iterator = run_command_with_exit_code(self.docker_compose_cmd + down_cmd, stream=stream)
        return iterator

    def start(
        self,
        services: Union[None, list[str]] = None,
        dry_run: bool = False,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()

        start_cmd: list[str] = ["start"]

        remove_parameters = ["services", "stream"]

        start_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        # doesn't checks if service exists or not
        if type(services) == list:
            start_cmd += services

        iterator = run_command_with_exit_code(self.docker_compose_cmd + start_cmd, stream=stream)
        return iterator

    def restart(
        self,
        services: Union[None, list[str]] = None,
        dry_run: bool = False,
        timeout: int = 100,
        no_deps: bool = False,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()

        parameters["timeout"] = str(timeout)

        restart_cmd: list[str] = ["restart"]

        remove_parameters = ["services", "stream"]

        restart_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        # doesn't checks if service exists or not
        if type(services) == list:
            restart_cmd += services

        iterator = run_command_with_exit_code(self.docker_compose_cmd + restart_cmd, stream=stream)
        return iterator

    def stop(
        self,
        services: Union[None, list[str]] = None,
        timeout: int = 100,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()
        parameters["timeout"] = str(timeout)

        stop_cmd: list[str] = ["stop"]

        remove_parameters = ["services", "stream"]

        stop_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        if type(services) == list:
            stop_cmd.extend(services)

        iterator = run_command_with_exit_code(self.docker_compose_cmd + stop_cmd, stream=stream)
        return iterator

    def exec(
        self,
        service: str,
        command: str,
        detach: bool = False,
        env: Union[None, list[str]] = None,
        no_tty: bool = False,
        privileged: bool = False,
        user: Union[None, str] = None,
        workdir: Union[None, str] = None,
        stream: bool = False,
        capture_output: bool = True,
        use_shlex_split: bool = True,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()

        exec_cmd: list[str] = ["exec"]

        remove_parameters = [
            "service",
            "stream",
            "command",
            "env",
            "use_shlex_split",
            "capture_output",
            "no_tty"
        ]

        exec_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        if type(env) == list:
            for i in env:
                exec_cmd += ["--env", i]

        exec_cmd += [service]

        if use_shlex_split:
            exec_cmd += shlex.split(command, posix=True)
        else:
            exec_cmd += [command]

        iterator = run_command_with_exit_code(
            self.docker_compose_cmd + exec_cmd, stream=stream, capture_output=capture_output
        )
        return iterator

    def ps(
        self,
        service: Union[None, list[str]] = None,
        dry_run: bool = False,
        all: bool = False,
        services: bool = False,
        filter: Union[
            None,
            Literal[
                "paused",
                "restarting",
                "removing",
                "running",
                "dead",
                "created",
                "exited",
            ],
        ] = None,
        format: Union[None, Literal["table", "json"]] = None,
        status: Union[
            None,
            list[
                Literal[
                    "paused",
                    "restarting",
                    "removing",
                    "running",
                    "dead",
                    "created",
                    "exited",
                ]
            ],
        ] = None,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()

        ps_cmd: list[str] = ["ps"]

        remove_parameters = [
            "service",
            "stream",
            "filter",
            "status",
        ]

        if filter:
            parameters["filter"] = f"status={filter}"

        ps_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        if type(status) == list:
            for i in status:
                ps_cmd += ["--status", i]

        if service:
            ps_cmd += service

        iterator = run_command_with_exit_code(self.docker_compose_cmd + ps_cmd, stream=stream)
        return iterator

    def logs(
        self,
        services: Union[None, list[str]] = None,
        dry_run: bool = False,
        follow: bool = False,
        no_color: bool = False,
        no_log_prefix: bool = False,
        since: Union[None, str] = None,
        tail: Union[None, int] = None,
        until: Union[None, int] = None,
        timestamps: bool = False,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()

        logs_cmd: list[str] = ["logs"]

        remove_parameters = ["services", "stream"]

        logs_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        if services:
            logs_cmd += services

        iterator = run_command_with_exit_code(
            self.docker_compose_cmd + logs_cmd,
            stream=stream,
        )
        return iterator

    def ls(
        self,
        all: bool = False,
        dry_run: bool = False,
        format: Literal["table", "json"] = "table",
    ):
        parameters: dict = locals()

        ls_cmd: list[str] = ["ls"]

        ls_cmd += parameters_to_options(parameters)

        try:
            output = run(self.docker_compose_cmd + ls_cmd, capture_output=True)
            output = output.stdout.decode()
        except:
            return False

        return output

    def pull(
        self,
        dry_run: bool = False,
        ignore_buildable: bool = False,
        ignore_pull_failures: bool = False,
        include_deps: bool = False,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()

        pull_cmd: list[str] = ["pull"]

        remove_parameters = ["stream"]

        pull_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        iterator = run_command_with_exit_code(
            self.docker_compose_cmd + pull_cmd,
            stream=stream,
        )
        return iterator

    def run(
        self,
        service: str,
        command: Optional[str] = None,
        name: Optional[str] = None,
        detach: bool = False,
        rm: bool = False,
        entrypoint: Optional[str] = None,
        use_shlex_split: bool = True,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()
        run_cmd: list = ["run"]

        remove_parameters = ["stream", "command", "service", "use_shlex_split"]

        run_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        run_cmd += [service]

        if command:
            if use_shlex_split:
                run_cmd += shlex.split(command, posix=True)
            else:
                run_cmd += [command]

        iterator = run_command_with_exit_code(self.docker_compose_cmd + run_cmd, stream=stream)
        return iterator

    def cp(
        self,
        source: str,
        destination: str,
        source_container: str = None,
        destination_container: str = None,
        archive: bool = False,
        follow_link: bool = False,
        stream: bool = False,
    ) -> Union[Iterable[Tuple[str, bytes]], SubprocessOutput]:
        parameters: dict = locals()

        cp_cmd: list = ["cp"]

        remove_parameters = [
            "stream",
            "source",
            "destination",
            "source_container",
            "destination_container",
        ]

        cp_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        if source_container:
            source = f"{source_container}:{source}"

        if destination_container:
            destination = f"{destination_container}:{destination}"

        cp_cmd += [f"{source}"]
        cp_cmd += [f"{destination}"]

        iterator = run_command_with_exit_code(self.docker_compose_cmd + cp_cmd, stream=stream)
        return iterator



from copy import deepcopy
from typing import Dict, List
import json
from pathlib import Path
import importlib.resources as pkg_resources

def get_frappe_manager_own_files(file_path: str):
    return Path(str(pkg_resources.files("frappe_manager").joinpath(file_path)))

def get_examples_from_toml(commands_stack: List[str], frappe_version: str, toml_path: Path = get_frappe_manager_own_files('./utils/examples.json')):
    file_data = toml_path.read_bytes()
    data: Dict[str,List[Dict[str,str]]] = json.loads(file_data)

    bench_name = 'example.com'

    example_data ={
        'benchname' : bench_name,
        'default_version' : frappe_version
    }

    examples_data = deepcopy(data)

    for command in commands_stack:
        if not command in examples_data:
            return None

        examples_data = examples_data[command]


    if not 'examples' in examples_data:
        return None

    examples_data = examples_data['examples']

    from rich.table import Table

    examples_table = Table(padding=(0,0),title=None,show_header=False,show_lines=False, box=None)

    if examples_data:
        element_example_data = deepcopy(example_data)

        for element in examples_data:
            desc = element.get('desc', 'None')
            code = element.get('code', 'None')

            if 'benchname' in element:
                element_example_data['benchname'] = element['benchname']

            element_table = Table(box=None,show_lines=False)

            element_table.add_row(f"[bold cyan]{desc.format(**example_data)}[/bold cyan]")
            element_table.add_row(f"[blue]:play_button:[/blue] fm {' '.join(commands_stack)} {element_example_data['benchname']}{code.format(**element_example_data)}")
            examples_table.add_row(element_table)
        return examples_table



VSCODE_LAUNCH_JSON = {
    "version": "0.2.0",
    "configurations": [
        {
            "name": "fm-frappe-debug",
            "type": "debugpy",
            "request": "launch",
            "program": "/workspace/frappe-bench/apps/frappe/frappe/utils/bench_helper.py",
            "args": ["frappe", "serve", "--port", "80", "--noreload", "--nothreading"],
            "cwd": "/workspace/frappe-bench/sites",
            "env": {"DEV_SERVER": "1"},
            "preLaunchTask": "fm-kill-port-80",
        }
    ],
}

VSCODE_TASKS_JSON = {
    "version": "2.0.0",
    "tasks": [
        {
            "label": "fm-kill-port-80",
            "type": "shell",
            "command": "/bin/bash",
            "args": ["-c", "fuser -k 80/tcp || true"],
            "presentation": {"reveal": "never", "panel": "dedicated"},
            "options": {"ignoreExitCode": True},
        }
    ],
}

VSCODE_SETTINGS_JSON = {
  "python.defaultInterpreterPath": "/workspace/frappe-bench/env/bin/python",
  "[python]": {
    "editor.defaultFormatter": "ms-python.black-formatter",
    "editor.detectIndentation": True,
    "editor.tabSize": 4,
    "editor.insertSpaces": True,
  },
  "black-formatter.importStrategy" : "useBundled"
}

PREBAKED_SITE_APPS= {
    "https://github.com/frappe/frappe": "version-15",
    "https://github.com/frappe/erpnext": "version-15",
    "https://github.com/frappe/hrms": "version-15",
}



from pathlib import Path
from typing import Tuple
from frappe_manager.ssl_manager.certificate import SSLCertificate
from frappe_manager.ssl_manager.ssl_certificate_service import SSLCertificateService
from frappe_manager.display_manager.DisplayManager import richprint


class NoOpCertificateService(SSLCertificateService):
    def __init__(self, root_dir: Path = Path('/dev/null')):
        self.root_dir = root_dir

    def renew_certificate(self):
        pass

    def remove_certificate(self, certificate: 'SSLCertificate'):
        richprint.warning(f"{certificate.domain} doesn't have certificate issued.")

    def generate_certificate(self, certificate: 'SSLCertificate') -> Tuple[Path, Path]:
        return Path('/dev/null'), Path('/dev/null')



from pathlib import Path
from frappe_manager.site_manager import bench_operations
from frappe_manager.site_manager.site_exceptions import BenchNotRunning
from frappe_manager.utils.site import pull_docker_images
import typer
import os
import sys
import shutil
from typing import Annotated, List, Optional
from frappe_manager.compose_project.compose_project import ComposeProject
from frappe_manager.services_manager.services_exceptions import ServicesNotCreated
from frappe_manager.site_manager.SiteManager import BenchesManager
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager import (
    CLI_BENCH_CONFIG_FILE_NAME,
    CLI_DIR,
    DEFAULT_EXTENSIONS,
    STABLE_APP_BRANCH_MAPPING_LIST,
    EnableDisableOptionsEnum,
    SiteServicesEnum,
    CLI_BENCHES_DIRECTORY,
)
from frappe_manager.docker_wrapper.DockerClient import DockerClient
from frappe_manager.logger import log
from frappe_manager.services_manager.services import ServicesManager
from frappe_manager.migration_manager.migration_executor import MigrationExecutor
from frappe_manager.site_manager.site import Bench
from frappe_manager.ssl_manager import LETSENCRYPT_PREFERRED_CHALLENGE, SUPPORTED_SSL_TYPES
from frappe_manager.ssl_manager.certificate import SSLCertificate
from frappe_manager.ssl_manager.letsencrypt_certificate import LetsencryptSSLCertificate
from frappe_manager.utils.callbacks import (
    apps_list_validation_callback,
    create_command_sitename_callback,
    frappe_branch_validation_callback,
    sites_autocompletion_callback,
    version_callback,
    sitename_callback,
    code_command_extensions_callback,
)
from frappe_manager.utils.helpers import (
    format_ssl_certificate_time_remaining,
    is_cli_help_called,
    get_current_fm_version,
)
from frappe_manager.services_manager.commands import services_root_command
from frappe_manager.sub_commands.self_commands import self_app
from frappe_manager.sub_commands.ssl_command import ssl_root_command
from frappe_manager.metadata_manager import FMConfigManager
from frappe_manager.site_manager.bench_config import BenchConfig, FMBenchEnvType
from frappe_manager.migration_manager.version import Version
from frappe_manager.compose_manager.ComposeFile import ComposeFile
from email_validator import validate_email

app = typer.Typer(no_args_is_help=True, rich_markup_mode="rich")
app.add_typer(services_root_command, name="services", help="Handle global services.")
app.add_typer(self_app, name="self", help="Perform operations related to the [bold][blue]fm[/bold][/blue] itself.")
app.add_typer(ssl_root_command, name="ssl", help="Perform operations related to ssl.")


@app.callback()
def app_callback(
    ctx: typer.Context,
    verbose: Annotated[bool, typer.Option("--verbose", "-v", help="Enable verbose output.")] = False,
    version: Annotated[
        Optional[bool], typer.Option("--version", "-V", help="Show Version.", callback=version_callback)
    ] = None,
):
    """
    Frappe-Manager for creating frappe development environments.
    """
    ctx.obj = {}
    help_called = is_cli_help_called(ctx)
    ctx.obj["is_help_called"] = help_called

    if not help_called:
        first_time_install = False

        richprint.start("Working")

        if not CLI_DIR.exists():
            # creating the sites dir
            # TODO check if it's writeable and readable -> by writing a file to it and catching exception
            CLI_DIR.mkdir(parents=True, exist_ok=True)
            CLI_BENCHES_DIRECTORY.mkdir(parents=True, exist_ok=True)
            richprint.print(f"fm directory doesn't exists! Created at -> {str(CLI_DIR)}")
            first_time_install = True
        else:
            if not CLI_DIR.is_dir():
                richprint.exit("Sites directory is not a directory! Aborting!")

        # logging
        global logger
        logger = log.get_logger()
        logger.info("")
        logger.info(f"{':'*20}FM Invoked{':'*20}")
        logger.info("")

        # logging command provided by user
        logger.info(f"RUNNING COMMAND: {' '.join(sys.argv[1:])}")
        logger.info("-" * 20)

        # check docker daemon service
        if not DockerClient().server_running():
            richprint.exit("Docker daemon not running. Please start docker service.")

        fm_config_manager: FMConfigManager = FMConfigManager.import_from_toml()

        # docker pull
        if first_time_install:
            if not fm_config_manager.root_path.exists():
                richprint.print("It seems like the first installation. Pulling docker images...️", "🔍")

                completed_status = pull_docker_images()

                if not completed_status:
                    shutil.rmtree(CLI_DIR)
                    richprint.exit("Aborting. Not able to pull all required Docker images.")

                current_version = Version(get_current_fm_version())
                fm_config_manager.version = current_version
                fm_config_manager.export_to_toml()

        migrations = MigrationExecutor(fm_config_manager)
        migration_status = migrations.execute()

        if not migration_status:
            richprint.exit(f"Rollbacked to previous version of fm {migrations.prev_version}.")

        services_manager: ServicesManager = ServicesManager(verbose=verbose)
        services_manager.set_typer_context(ctx)

        services_manager.init()

        try:
            services_manager.entrypoint_checks(start=True)
        except ServicesNotCreated as e:
            services_manager.remove_itself()
            richprint.exit(f"Not able to create services. {e}")

        ctx.obj["services"] = services_manager
        ctx.obj["verbose"] = verbose
        ctx.obj['fm_config_manager'] = fm_config_manager


@app.command(no_args_is_help=True)
def create(
    ctx: typer.Context,
    benchname: Annotated[str, typer.Argument(help="Name of the bench", callback=create_command_sitename_callback)],
    apps: Annotated[
        List[str],
        typer.Option(
            "--apps",
            "-a",
            help="FrappeVerse apps to install. App should be specified in format <appname>:<branch> or <appname>.",
            callback=apps_list_validation_callback,
            show_default=False,
        ),
    ] = [],
    environment: Annotated[
        FMBenchEnvType, typer.Option("--environment", "-e", help="Select bench environment type.")
    ] = FMBenchEnvType.dev,
    letsencrypt_preferred_challenge: Annotated[
        Optional[LETSENCRYPT_PREFERRED_CHALLENGE],
        typer.Option(help="Select preferred letsencrypt challenge.", show_default=False),
    ] = None,
    letsencrypt_email: Annotated[
        Optional[str],
        typer.Option(help="Specify email for letsencrypt", show_default=False),
    ] = None,
    developer_mode: Annotated[
        EnableDisableOptionsEnum, typer.Option(help="Toggle frappe developer mode.")
    ] = EnableDisableOptionsEnum.disable,
    frappe_branch: Annotated[
        str, typer.Option(help="Specify the branch name for frappe app", callback=frappe_branch_validation_callback)
    ] = "version-15",
    template: Annotated[bool, typer.Option(help="Create template bench.")] = False,
    admin_pass: Annotated[
        str,
        typer.Option(
            help="Default Password for the standard 'Administrator' User. This will be used as the password for the Administrator User for all new bench."
        ),
    ] = "admin",
    ssl: Annotated[
        SUPPORTED_SSL_TYPES, typer.Option(help="Enable https", show_default=True)
    ] = SUPPORTED_SSL_TYPES.none,
):
    # TODO Create markdown table for the below help
    """
    Create a new bench.
    """

    services_manager: ServicesManager = ctx.obj["services"]
    fm_config_manager: FMConfigManager = ctx.obj["fm_config_manager"]
    verbose = ctx.obj['verbose']

    benches = BenchesManager(CLI_BENCHES_DIRECTORY, services=services_manager, verbose=verbose)
    benches.set_typer_context(ctx)

    bench_path = benches.root_path / benchname
    bench_config_path = bench_path / CLI_BENCH_CONFIG_FILE_NAME

    if ssl == SUPPORTED_SSL_TYPES.le:
        if not letsencrypt_preferred_challenge:
            if fm_config_manager.letsencrypt.exists:
                if letsencrypt_preferred_challenge is None:
                    letsencrypt_preferred_challenge = LETSENCRYPT_PREFERRED_CHALLENGE.dns01

            if not letsencrypt_preferred_challenge:
                letsencrypt_preferred_challenge = LETSENCRYPT_PREFERRED_CHALLENGE.http01

        if fm_config_manager.letsencrypt.email == 'dummy@fm.fm' or fm_config_manager.letsencrypt.email is None:
            if not letsencrypt_email:
                richprint.stop()
                raise typer.BadParameter("No email provided, required by certbot.", param_hint='--letsencrypt-email')
            else:
                email = letsencrypt_email

            validate_email(email, check_deliverability=False)
        else:
            richprint.print(
                "Defaulting to Let's Encrypt email from [blue]fm_config.toml[/blue] since [blue]'--letsencrypt-email'[/blue] is not given."
            )
            email = fm_config_manager.letsencrypt.email

        ssl_certificate = LetsencryptSSLCertificate(
            domain=benchname,
            ssl_type=ssl,
            email=email,
            preferred_challenge=letsencrypt_preferred_challenge,
            api_key=fm_config_manager.letsencrypt.api_key,
            api_token=fm_config_manager.letsencrypt.api_token,
        )

    elif ssl == SUPPORTED_SSL_TYPES.none:
        ssl_certificate = SSLCertificate(domain=benchname, ssl_type=ssl)

    if developer_mode == EnableDisableOptionsEnum.enable:
        developer_mode_status = True
    elif developer_mode == EnableDisableOptionsEnum.disable:
        developer_mode_status = False

    bench_config: BenchConfig = BenchConfig(
        name=benchname,
        apps_list=apps,
        frappe_branch=frappe_branch,
        developer_mode=True if environment == FMBenchEnvType.dev else developer_mode_status,
        admin_tools=True if environment == FMBenchEnvType.dev else False,
        admin_pass=admin_pass,
        # TODO get this info from services, maybe ?
        environment_type=environment,
        root_path=bench_config_path,
        ssl=ssl_certificate,
    )

    compose_path = bench_path / 'docker-compose.yml'
    compose_file_manager = ComposeFile(compose_path)
    compose_project = ComposeProject(compose_file_manager, verbose)

    bench: Bench = Bench(bench_path, benchname, bench_config, compose_project, services_manager)
    benches.add_bench(bench)
    benches.create_benches(is_template_bench=template)


@app.command()
def delete(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
    force: Annotated[bool, typer.Option("--force", "-f", help="Force delete bench.")] = False,
):
    """Delete a bench."""

    if benchname:
        services_manager = ctx.obj["services"]
        verbose = ctx.obj['verbose']
        benches = BenchesManager(CLI_BENCHES_DIRECTORY, services=services_manager, verbose=verbose)
        benches.set_typer_context(ctx)

        bench_path: Path = benches.root_path / benchname
        bench_compose_path = bench_path / 'docker-compose.yml'
        compose_file_manager = ComposeFile(bench_compose_path)
        bench_compose_project = ComposeProject(compose_file_manager)

        bench_config_path = bench_path / CLI_BENCH_CONFIG_FILE_NAME
        # try using bench object if not then create bench

        if not bench_config_path.exists():
            uid: int = os.getuid()
            gid: int = os.getgid()

            # generate fake bench
            fake_config = BenchConfig(
                name=benchname,
                userid=uid,
                usergroup=gid,
                apps_list=[],
                frappe_branch=STABLE_APP_BRANCH_MAPPING_LIST['frappe'],
                developer_mode=False,
                # TODO do something about this forcefully delete maybe
                admin_tools=False,
                admin_pass='pass',
                environment_type=FMBenchEnvType.dev,
                ssl=SSLCertificate(domain=benchname, ssl_type=SUPPORTED_SSL_TYPES.none),
                root_path=bench_config_path,
            )

            bench = Bench(
                bench_path,
                benchname,
                fake_config,
                bench_compose_project,
                services=services_manager,
                workers_check=False,
            )

        else:
            bench = Bench.get_object(benchname, services=services_manager, workers_check=False, admin_tools_check=False)

        benches.add_bench(bench)
        benches.remove_benches()


@app.command()
def list(ctx: typer.Context):
    """Lists all of the available benches."""

    services_manager = ctx.obj["services"]
    verbose = ctx.obj['verbose']
    benches = BenchesManager(CLI_BENCHES_DIRECTORY, services=services_manager, verbose=verbose)
    benches.set_typer_context(ctx)
    benches.list_benches()


@app.command()
def start(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
    force: Annotated[bool, typer.Option("--force", "-f", help="Force recreate bench containers")] = False,
):
    """Start a bench."""

    services_manager = ctx.obj["services"]
    verbose = ctx.obj['verbose']
    bench = Bench.get_object(benchname, services_manager)
    bench.start(force=force)


@app.command()
def stop(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
):
    """Stop a bench."""

    services_manager = ctx.obj["services"]
    verbose = ctx.obj['verbose']
    bench = Bench.get_object(benchname, services_manager)
    benches = BenchesManager(CLI_BENCHES_DIRECTORY, services=services_manager, verbose=verbose)
    benches.add_bench(bench)
    benches.stop_benches()


@app.command()
def code(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
    user: Annotated[str, typer.Option(help="Connect as this user.")] = "frappe",
    extensions: Annotated[
        List[str],
        typer.Option(
            "--extension",
            "-e",
            help="List of extensions to install in vscode at startup.Provide extension id eg: ms-python.python",
            callback=code_command_extensions_callback,
            show_default=False,
        ),
    ] = DEFAULT_EXTENSIONS,
    force_start: Annotated[
        bool, typer.Option("--force-start", "-f", help="Force start the site before attaching to container.")
    ] = False,
    debugger: Annotated[bool, typer.Option("--debugger", "-d", help="Sync vscode debugger configuration.")] = False,
    workdir: Annotated[
        str, typer.Option("--work-dir", "-w", help="Set working directory in vscode.")
    ] = '/workspace/frappe-bench',
):
    """Open bench in vscode."""

    services_manager = ctx.obj["services"]
    verbose = ctx.obj['verbose']
    bench = Bench.get_object(benchname, services_manager)

    if force_start:
        bench.start()

    bench.attach_to_bench(user=user, extensions=extensions, workdir=workdir, debugger=debugger)


@app.command()
def logs(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
    service: Annotated[
        Optional[SiteServicesEnum], typer.Option(help="Specify compose service name to show container logs.")
    ] = None,
    follow: Annotated[bool, typer.Option("--follow", "-f", help="Follow logs.")] = False,
):
    """Show frappe server logs or container logs for a given bench."""

    services_manager = ctx.obj["services"]
    verbose = ctx.obj['verbose']
    bench = Bench.get_object(benchname, services_manager)
    bench.logs(follow, service)


@app.command()
def shell(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
    user: Annotated[Optional[str], typer.Option(help="Connect as this user.", show_default=False)] = None,
    service: Annotated[
        SiteServicesEnum, typer.Option(help="Specify compose service name for which to spawn shell.")
    ] = SiteServicesEnum.frappe,
):
    """Spawn shell for the give bench."""

    services_manager = ctx.obj["services"]
    verbose = ctx.obj['verbose']
    bench = Bench.get_object(benchname, services_manager)
    bench.shell(SiteServicesEnum(service).value, user)


@app.command()
def info(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
):
    """Shows information about given bench."""

    services_manager = ctx.obj["services"]
    verbose = ctx.obj['verbose']
    bench = Bench.get_object(benchname, services_manager)
    bench.info()


@app.command()
def update(
    ctx: typer.Context,
    benchname: Annotated[
        Optional[str],
        typer.Argument(
            help="Name of the bench.", autocompletion=sites_autocompletion_callback, callback=sitename_callback
        ),
    ] = None,
    ssl: Annotated[Optional[SUPPORTED_SSL_TYPES], typer.Option(help="Enable SSL.", show_default=False)] = None,
    admin_tools: Annotated[
        Optional[EnableDisableOptionsEnum],
        typer.Option("--admin-tools", help="Toggle admin-tools.", show_default=False),
    ] = None,
    letsencrypt_preferred_challenge: Annotated[
        Optional[LETSENCRYPT_PREFERRED_CHALLENGE],
        typer.Option(help="Select preferred letsencrypt challenge.", show_default=False),
    ] = None,
    letsencrypt_email: Annotated[
        Optional[str],
        typer.Option(help="Specify email for letsencrypt", show_default=False),
    ] = None,
    environment: Annotated[
        Optional[FMBenchEnvType],
        typer.Option("--environment", "-e", help="Switch bench environment.", show_default=False),
    ] = None,
    developer_mode: Annotated[
        Optional[EnableDisableOptionsEnum],
        typer.Option(help="Toggle frappe developer mode.", show_default=False),
    ] = None,
):
    """Update bench."""

    services_manager = ctx.obj["services"]
    bench = Bench.get_object(benchname, services_manager)
    fm_config_manager: FMConfigManager = ctx.obj["fm_config_manager"]

    restart_required = False
    bench_config_save = False

    if not bench.compose_project.running:
        raise BenchNotRunning(bench_name=bench.name)

    if developer_mode:
        if developer_mode == EnableDisableOptionsEnum.enable:
            bench.bench_config.developer_mode = True
            richprint.print("Enabling frappe developer mode.")
            bench.common_bench_config_set({'developer_mode': bench.bench_config.developer_mode})
            richprint.print("Enabled frappe developer mode.")
        elif developer_mode == EnableDisableOptionsEnum.disable:
            bench.bench_config.developer_mode = False
            richprint.print("Disabling frappe developer mode.")
            bench.common_bench_config_set({'developer_mode': bench.bench_config.developer_mode})
            richprint.print("Enabled frappe developer mode.")

        bench_config_save = True
        restart_required = True

    if environment:
        richprint.change_head(f"Switching bench environemnt to {environment.value}")
        bench.bench_config.environment_type = environment
        bench.switch_bench_env()
        richprint.print(f"Switched bench environemnt to {environment.value}.")
        bench_config_save = True

    if ssl:
        new_ssl_certificate = SSLCertificate(domain=benchname, ssl_type=SUPPORTED_SSL_TYPES.none)

        if ssl == SUPPORTED_SSL_TYPES.le:
            if not letsencrypt_preferred_challenge:
                if fm_config_manager.letsencrypt.exists:
                    if letsencrypt_preferred_challenge is None:
                        letsencrypt_preferred_challenge = LETSENCRYPT_PREFERRED_CHALLENGE.dns01

                if not letsencrypt_preferred_challenge:
                    letsencrypt_preferred_challenge = LETSENCRYPT_PREFERRED_CHALLENGE.http01

            if fm_config_manager.letsencrypt.email == 'dummy@fm.fm' or fm_config_manager.letsencrypt.email is None:
                if not letsencrypt_email:
                    richprint.stop()
                    raise typer.BadParameter(
                        "No email provided, required by certbot.", param_hint='--letsencrypt-email'
                    )
                else:
                    email = letsencrypt_email

                validate_email(email, check_deliverability=False)
            else:
                richprint.print(
                    "Defaulting to Let's Encrypt email from [blue]fm_config.toml[/blue] since [blue]'--letsencrypt-email'[/blue] is not given."
                )
                email = fm_config_manager.letsencrypt.email

            new_ssl_certificate = LetsencryptSSLCertificate(
                domain=benchname,
                ssl_type=ssl,
                email=email,
                preferred_challenge=letsencrypt_preferred_challenge,
                api_key=fm_config_manager.letsencrypt.api_key,
                api_token=fm_config_manager.letsencrypt.api_token,
            )

        richprint.print("Updating Certificate.")
        bench.update_certificate(new_ssl_certificate)
        richprint.print("Certificate Updated.")

        if bench.has_certificate():
            richprint.print(
                f"SSL Certificate will expire in {format_ssl_certificate_time_remaining(bench.certificate_manager.get_certficate_expiry())}"
            )

    if admin_tools:
        restart_required = False
        if admin_tools == EnableDisableOptionsEnum.enable:
            richprint.change_head("Enabling Admin-tools")
            bench.bench_config.admin_tools = True
            if not bench.admin_tools.compose_project.compose_file_manager.compose_path.exists():
                restart_required = bench.sync_admin_tools_compose()
            else:
                restart_required = bench.admin_tools.enable()
            bench_config_save = True
            richprint.print("Enabled Admin-tools.")

        elif admin_tools == EnableDisableOptionsEnum.disable:
            if (
                not bench.admin_tools.compose_project.compose_file_manager.compose_path.exists()
                or not bench.bench_config.admin_tools
            ):
                richprint.print("Admin tools is already disabled.")
                return
            else:
                bench.bench_config.admin_tools = False
                restart_required = bench.admin_tools.disable()
                bench_config_save = True

    # prompt for restart frappe server
    if restart_required:
        should_restart = richprint.prompt_ask(
            prompt=f"Frappe server restart is required after {admin_tools.value} of admin tools. Do you want to proceed ?",
            choices=['yes', 'no'],
        )
        if should_restart == 'yes':
            bench.restart_frappe_server()

    if bench_config_save:
        bench.save_bench_config()



from pathlib import Path
from frappe_manager.migration_manager.migration_base import MigrationBase
from frappe_manager.migration_manager.migration_exections import MigrationExceptionInBench
from frappe_manager.migration_manager.migration_helpers import (
    MigrationBench,
    MigrationBenches,
    MigrationServicesManager,
)
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.migration_manager.version import Version
from frappe_manager.migration_manager.backup_manager import BackupManager


class MigrationV0110(MigrationBase):
    version = Version("0.11.0")

    def init(self):
        self.cli_dir: Path = Path.home() / 'frappe'
        self.benches_dir = self.cli_dir / "sites"
        self.backup_manager = BackupManager(name=str(self.version), benches_dir=self.benches_dir)
        self.benches_manager = MigrationBenches(self.benches_dir)
        self.services_manager: MigrationServicesManager = MigrationServicesManager(
            services_path=self.cli_dir / 'services'
        )

    def migrate_bench(self, bench: MigrationBench):
        bench.compose_project.down_service(volumes=False)
        self.migrate_bench_compose(bench)

    def migrate_bench_compose(self, bench: MigrationBench):
        richprint.change_head("Migrating bench compose")

        if not bench.compose_project.compose_file_manager.exists():
            richprint.print(f"Failed to migrate {bench.name} compose file.")
            raise MigrationExceptionInBench(f"{bench.compose_project.compose_file_manager.compose_path} not found.")

        # change image tag to the latest
        # in this migration only tag of frappe container is changed
        images_info = bench.compose_project.compose_file_manager.get_all_images()
        image_info = images_info['frappe']

        # get v0.11.0 frappe image
        image_info['tag'] = self.version.version_string()
        image_info['name'] = 'ghcr.io/rtcamp/frappe-manager-frappe'

        output = bench.compose_project.docker.pull(
            container_name=f"{image_info['name']}:{image_info['tag']}", stream=True
        )
        richprint.live_lines(output, padding=(0, 0, 0, 2))

        bench.compose_project.compose_file_manager.set_all_images(images_info)

        bench.compose_project.compose_file_manager.set_version(str(self.version))
        bench.compose_project.compose_file_manager.write_to_file()

        richprint.print(f"Migrated [blue]{bench.name}[/blue] compose file.")



from enum import Enum


class SUPPORTED_SSL_TYPES(str, Enum):
    le = 'letsencrypt'
    none = 'disable'


class LETSENCRYPT_PREFERRED_CHALLENGE(str, Enum):
    dns01 = 'dns01'
    http01 = 'http01'



import json
import shlex
from sys import exception

from typing import Literal, Optional, List
from pathlib import Path
from frappe_manager.docker_wrapper.DockerCompose import DockerComposeWrapper
from frappe_manager.display_manager.DisplayManager import richprint
from frappe_manager.docker_wrapper.DockerException import DockerException
from frappe_manager.utils.docker import (
    SubprocessOutput,
    is_current_user_in_group,
    parameters_to_options,
    run_command_with_exit_code,
)


class DockerClient:
    """
    This class provide one to one mapping to the docker command.

    Only this args have are different use case.
        stream (bool, optional): A boolean flag indicating whether to stream the output of the command as it runs.
            If set to True, the output will be displayed in real-time. If set to False, the output will be
            displayed after the command completes. Defaults to False.
        stream_only_exit_code (bool, optional): A boolean flag indicating whether to only stream the exit code of the
            command. Defaults to False.
    """

    def __init__(self, compose_file_path: Optional[Path] = None):
        """
        Initializes a DockerClient object.
        Args:
            compose_file_path (Optional[Path]): The path to the Docker Compose file. Defaults to None.
        """
        self.docker_cmd = ["docker"]
        if compose_file_path:
            self.compose = DockerComposeWrapper(compose_file_path)

    def version(self) -> dict:
        """
        Retrieves the version information of the Docker client.

        Returns:
            A dictionary containing the version information.
        """
        parameters: dict = locals()

        parameters["format"] = "json"

        ver_cmd: list = ["version"]

        ver_cmd += parameters_to_options(parameters)

        try:
            output: SubprocessOutput = run_command_with_exit_code(self.docker_cmd + ver_cmd, stream=False)
            version: dict = json.loads(" ".join(output.stdout))
            return version
        except DockerException as e:
            version: dict = json.loads(" ".join(e.output.stdout))
            return version

    def server_running(self) -> bool:
        """
        Checks if the Docker server is running.

        Returns:
            bool: True if the Docker server is running, False otherwise.
        """
        docker_info = self.version()

        if "Server" in docker_info:
            if docker_info['Server']:
                return True
            else:
                return False
        else:
            # check if the current user in the docker group and notify the user
            is_current_user_in_group("docker")
            return False

    def cp(
        self,
        source: str,
        destination: str,
        source_container: Optional[str] = None,
        destination_container: Optional[str] = None,
        archive: bool = False,
        follow_link: bool = False,
        stream: bool = False,
    ):
        parameters: dict = locals()
        cp_cmd: list = ["cp"]

        remove_parameters = [
            "stream",
            "source",
            "destination",
            "source_container",
            "destination_container",
        ]

        cp_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        if source_container:
            source = f"{source_container}:{source}"

        if destination_container:
            destination = f"{destination_container}:{destination}"

        cp_cmd += [f"{source}"]
        cp_cmd += [f"{destination}"]

        iterator = run_command_with_exit_code(self.docker_cmd + cp_cmd, stream=stream)
        return iterator

    def kill(
        self,
        container: str,
        signal: Optional[str] = None,
        stream: bool = False,
    ):
        parameters: dict = locals()
        kill_cmd: list = ["kill"]

        remove_parameters = ["stream", "container"]

        kill_cmd += parameters_to_options(parameters, exclude=remove_parameters)
        kill_cmd += [f"{container}"]

        iterator = run_command_with_exit_code(self.docker_cmd + kill_cmd, stream=stream)
        return iterator

    def rm(
        self,
        container: str,
        force: bool = False,
        link: bool = False,
        volumes: bool = False,
        stream: bool = False,
    ):
        parameters: dict = locals()
        rm_cmd: list = ["rm"]

        remove_parameters = ["stream", "container"]

        rm_cmd += parameters_to_options(parameters, exclude=remove_parameters)
        rm_cmd += [f"{container}"]

        iterator = run_command_with_exit_code(self.docker_cmd + rm_cmd, stream=stream)
        return iterator

    def run(
        self,
        image: str,
        command: Optional[str] = None,
        env: Optional[List[str]] = None,
        name: Optional[str] = None,
        volume: Optional[str] = None,
        detach: bool = False,
        entrypoint: Optional[str] = None,
        pull: Literal["missing", "never", "always"] = "missing",
        use_shlex_split: bool = True,
        stream: bool = False,
    ):
        parameters: dict = locals()
        run_cmd: list = ["run"]

        remove_parameters = ["stream", "command", "image", "use_shlex_split", "env"]

        run_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        if isinstance(env, list):
            for i in env:
                run_cmd += ["--env", i]

        run_cmd += [f"{image}"]

        if command:
            if use_shlex_split:
                run_cmd += shlex.split(command, posix=True)
            else:
                run_cmd += [command]

        iterator = run_command_with_exit_code(self.docker_cmd + run_cmd, stream=stream)
        return iterator

    def pull(
        self,
        container_name: str,
        all_tags: bool = False,
        platform: Optional[str] = None,
        stream: bool = False,
    ):
        parameters: dict = locals()

        pull_cmd: list[str] = ["pull"]

        remove_parameters = ["stream", "container_name"]

        pull_cmd += parameters_to_options(parameters, exclude=remove_parameters)
        pull_cmd += [container_name]

        iterator = run_command_with_exit_code(
            self.docker_cmd + pull_cmd,
            stream=stream,
        )
        return iterator

    def images(
        self,
        format: Literal['json'] = 'json',
    ):
        parameters: dict = locals()

        images_cmd: list[str] = ["images"]
        remove_parameters = []

        images_cmd += parameters_to_options(parameters, exclude=remove_parameters)

        output: SubprocessOutput = run_command_with_exit_code(
            self.docker_cmd + images_cmd,
            stream=False,
        )

        images = []

        if output.stdout:
            for image in output.stdout:
                images.append(json.loads(image))

        return images



import shutil
from datetime import datetime
from pathlib import Path
from frappe_manager import CLI_BENCHES_DIRECTORY, CLI_DIR
from dataclasses import dataclass
from typing import Optional
from frappe_manager.logger import log

random_strings = []


@dataclass
class BackupData:
    src: Path
    dest: Path
    bench: Optional[str] = None
    bench_path: Optional[Path] = None
    prefix_timestamp: bool = False
    allow_restore: bool = True
    _is_restored: bool = False
    _prefix_length: int = 5

    @property
    def is_restored(self) -> bool:
        return self._is_restored

    @is_restored.setter
    def is_restored(self, v: bool) -> None:
        self._is_restored = v

    def __post_init__(self):
        file_name = self.dest.name

        if self.prefix_timestamp:
            while True:
                # Get current date and time
                now = datetime.now()
                # Format date and time
                current_time = now.strftime('%d-%b-%y--%H-%M-%S')

                if current_time not in random_strings:
                    random_strings.append(current_time)
                    file_name = f"{self.dest.name}-{current_time}"
                    break

        self.real_dest = self.dest.parent
        self.real_dest: Path = self.real_dest / file_name

    def exists(self):
        return self.dest.exists()


current_migration_timestamp = f"{datetime.now().strftime('%d-%b-%y--%H-%M-%S')}"
CLI_MIGARATIONS_DIR = CLI_DIR / 'backups'


class BackupManager:
    def __init__(
        self,
        name: str,
        backup_group_name: str = 'migrations',
        benches_dir: Path = CLI_BENCHES_DIRECTORY,
        backup_dir: Path = CLI_MIGARATIONS_DIR,
    ):
        self.name = name
        self.backup_group_name = backup_group_name
        self.root_backup_dir: Path = backup_dir / backup_group_name / current_migration_timestamp
        self.benches_dir: Path = benches_dir
        self.backup_dir: Path = self.root_backup_dir / self.name
        self.bench_backup_dir: Path = Path('backups') / backup_group_name / current_migration_timestamp
        self.backups = []
        self.logger = log.get_logger()

        # create backup dir if not exists
        self.backup_dir.mkdir(parents=True, exist_ok=True)

    def backup(
        self,
        src: Path,
        dest: Optional[Path] = None,
        bench_name: Optional[str] = None,
        allow_restore: bool = True,
    ):
        if not src.exists():
            return None

        if not dest:
            dest = self.backup_dir / src.name

            if bench_name:
                dest: Path = self.benches_dir / bench_name / self.bench_backup_dir / self.name / src.name

                if self.name == self.backup_group_name:
                    dest: Path = self.benches_dir / bench_name / self.bench_backup_dir / src.name

        backup_data = BackupData(src, dest, bench=bench_name, allow_restore=allow_restore)

        if not backup_data.real_dest.parent.exists():
            backup_data.real_dest.parent.mkdir(parents=True, exist_ok=True)

        self.logger.debug(f"Backup: {backup_data.src} => {backup_data.real_dest} ")

        if src.is_dir():
            # Copy directory
            shutil.copytree(backup_data.src, backup_data.real_dest)
        else:
            # Copy file
            shutil.copy2(backup_data.src, backup_data.real_dest)

        self.backups.append(backup_data)

        return backup_data

    def restore(self, backup_data, force=False):
        """
        Restore a file from a backup.
        """
        if not backup_data.allow_restore:
            return None

        if not backup_data.real_dest.exists():
            # print(f"No backup found at {backup_data.real_dest}")
            return None

        if force:
            self.logger.debug(f"Restore: {backup_data.real_dest} => {backup_data.src} ")
            if backup_data.src.exists():
                if backup_data.src.is_dir():
                    shutil.rmtree(backup_data.src)
                else:
                    backup_data.src.unlink()

        dest = shutil.copy(backup_data.real_dest, backup_data.src)

        backup_data.is_restored = True

        return dest
        # print(f"Restored {backup_data.src} from backup")

    def delete(self, backup_data):
        """
        Delete a specific backup.
        """
        if not backup_data.real_dest.exists():
            # print(f"No backup found at {backup_data.real_dest}")
            return None

        shutil.rmtree(backup_data.real_dest)

        self.backups.remove(backup_data)
        # print(f"Deleted backup at {backup_data.real_dest}")

    def delete_all(self):
        """
        Delete all backups.
        """
        for backup_data in self.backups:
            if backup_data.real_dest.exists():
                shutil.rmtree(backup_data.real_dest)
                # print(f"Deleted backup at {backup_data.real_dest}")

        self.backups.clear()
        # print("Deleted all backups")
